# -*- coding: utf-8 -*-
# @Time    : 2022/7/7 11:01 AM
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : prettyPrint.py
import json
from typing import List
from treelib import Tree
from prettytable import PrettyTable

from ultipa.printer.treeNode import TreeNode
from ultipa.types.types import DataItem, ExplainPlan, ResultType, Path, Node, Edge, PlanNode
from ultipa.utils.convert import convertTableToDict

class PrettyPrint:
	'''
	    Class that defines settings for formatted output
	'''
	count = 34
	tree = Tree()

	def printLine(self, name):
		num = int(PrettyPrint.count - len(name) / 2)
		print('-' * num + f'{name}' + '-' * num)

	def prettyStatus(self, data):
		y = PrettyTable(['Code', 'Message'], title="STATUS", end=10, align="l")
		y.add_row([data.code, data.message])
		print(y)

	def prettyStatistics(self, data):
		y = PrettyTable(['edgeAffected', 'nodeAffected', 'engineCost', 'totalCost'], title="Statistics", end=10,
						align="l")
		y.add_row([data.edgeAffected, data.nodeAffected, data.engineCost, data.totalCost])
		print(y)

	def prettyDataList(self, data: List[dict]):
		propertList = []
		if len(data) > 0:
			if len(data[0].keys()) > 0:
				propertList.extend(list(data[0].keys()))
			y = PrettyTable(propertList, title=f"Alias: Data", align="l")
			for d in data:
				li = []
				for p in propertList:
					li.append(d.get(p))
				y.add_row(li)
			print(y)
		else:
			return None

	def prettyData(self, data: dict):
		propertList = []
		if len(data.keys()) > 0:
			propertList.extend(list(data.keys()))
		y = PrettyTable(propertList, title=f"Alias: Data", align="l")
		li = []
		for p in propertList:
			li.append(data.get(p))
		y.add_row(li)
		print(y)

	def prettyNode(self, data: DataItem):
		schemaDict = {}
		for node in data.asNodes():
			propertList = ["id", "uuid", "schema"]
			if node is None:
				schemaDict.update({None: propertList})
				continue
			if not schemaDict.get(node.getSchema()):
				if node.values:
					newPropertList = propertList + list(node.values.keys())
					schemaDict.update({node.getSchema(): newPropertList})
				else:
					schemaDict.update({node.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {data.alias} AliasType: {data.type} Schema: {key}",
							align="l")
			for newNode in data.asNodes():
				if newNode is None:
					li = [None, None,None]
					y.add_row(li)
					continue
				if newNode.getSchema() == key:
					li = [newNode.getID(), newNode.getUUID(), newNode.getSchema()]
					if newNode.values:
						li.extend(list(newNode.values.values()))
					y.add_row(li)
				else:
					continue
			print(y)

	def prettyAttrNode(self, alias, nodes: List[Node]):
		ret = []
		schemaDict = {}
		for node in nodes:
			propertList = ["id", "uuid", "schema"]
			if node is None:
				schemaDict.update({None: propertList})
				continue
			if not schemaDict.get(node.getSchema()):
				if node.values:
					newPropertList = propertList + list(node.values.keys())
					schemaDict.update({node.getSchema(): newPropertList})
				else:
					schemaDict.update({node.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {alias} AliasType: NODE Schema: {key}",
							align="l")
			for newNode in nodes:
				if newNode is None:
					li = [None, None,None]
					y.add_row(li)
					continue
				if newNode.getSchema() == key:
					li = [newNode.getID(), newNode.getUUID(), newNode.getSchema()]
					if newNode.values:
						li.extend(list(newNode.values.values()))
					y.add_row(li)
				else:
					continue
			ret.append(y)
		return ret

	def prettyEdge(self, data: DataItem):
		schemaDict = {}
		for edge in data.asEdges():
			propertList = ["uuid", "from_uuid", "to_uuid", "from_id", "to_id", "schema"]
			if edge is None:
				schemaDict.update({None: propertList})
				continue
			if not schemaDict.get(edge.getSchema()):
				if edge.values:
					newPropertList = propertList + list(edge.values.keys())
					schemaDict.update({edge.getSchema(): newPropertList})
				else:
					schemaDict.update({edge.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {data.alias} AliasType: {data.type} Schema: {key}",
							align="l")
			for newEdge in data.asEdges():
				if newEdge is None:
					li = [None, None,None,None, None,None]
					y.add_row(li)
					continue
				if newEdge.getSchema() == key:
					li = [newEdge.getUUID(), newEdge.getFromUUID(), newEdge.getToUUID(), newEdge.getFrom(),
						  newEdge.getTo(), newEdge.getSchema()]
					if newEdge.values:
						li.extend(list(newEdge.values.values()))
					y.add_row(li)
				else:
					continue
			print(y)

	def prettyAttrEdge(self, alias, edges: List[Edge]):
		ret = []
		schemaDict = {}
		for edge in edges:
			propertList = ["uuid", "from_uuid", "to_uuid", "from_id", "to_id", "schema"]
			if edge is None:
				schemaDict.update({None: propertList})
				continue
			if not schemaDict.get(edge.getSchema()):
				if edge.values:
					newPropertList = propertList + list(edge.values.keys())
					schemaDict.update({edge.getSchema(): newPropertList})
				else:
					schemaDict.update({edge.getSchema(): propertList})

		for key in schemaDict:
			y = PrettyTable(schemaDict.get(key), title=f"Alias: {alias} AliasType: EDGE Schema: {key}", align="l")
			for newEdge in edges:
				if newEdge is None:
					li = [None, None,None,None, None,None]
					y.add_row(li)
					continue
				if newEdge.getSchema() == key:
					li = [newEdge.getUUID(), newEdge.getFromUUID(), newEdge.getToUUID(), newEdge.getFrom(),
						  newEdge.getTo(), newEdge.getSchema()]
					if newEdge.values:
						li.extend(list(newEdge.values.values()))
					y.add_row(li)
				else:
					continue
			ret.append(y)
		return ret

	def prettyTable(self, data: DataItem):

		if data.alias == "_algoList":
			headers = data.asTable().getHeaders()
			rows = data.asTable().getRows()
			table_rows_dict = convertTableToDict(rows, headers)
			algo_list = []
			for data1 in table_rows_dict:
				algo_list.append(json.loads(data1.get("param")))
			if len(algo_list) > 0:
				propertList = algo_list[0].keys()
				y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
				for algo in algo_list:
					row = []
					for key in propertList:
						row.append(algo.get(key))
					y.add_row(row)
			else:
				return
		else:
			propertList = []
			if len(data.asTable().rows) > 0:
				for index, header in enumerate(data.asTable().getHeaders()):
					propertList.append(header.get("property_name"))
			else:
				return

			y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
			for row in data.asTable().getRows():
				y.add_row(row)
		print(y)

	def _PrintPath(self, path: Path):
		pathStr = ""
		end = ""
		for index, edge in enumerate(path.edges):
			d1 = "-"
			d2 = "-"
			node = path.getNodes()[index]
			if edge.getFrom() == node.getID():
				d2 = "->"
				if index == len(path.edges) - 1:
					end = f" ({str(edge.getTo())})"
			else:
				d1 = "<-"
				if index == len(path.edges) - 1:
					end = f" ({str(edge.getFrom())})"

			pathStr += f"({node.getID()}) {d1} [{edge.getUUID()}] {d2}"
			if end:
				pathStr += end
		return pathStr

	def prettyPath(self, data: DataItem):
		propertList = ["#", "Path"]
		y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
		for i, path in enumerate(data.asPaths()):
			li = [i]
			pathStr = self._PrintPath(path)
			li.append(pathStr)
			y.add_row(li)
		print(y)

	def _formatAttrListData(self, data: list, alias: str):
		if isinstance(data,list) and len(data) > 0:
			if isinstance(data[0], Path):
				li = []
				for i in data:
					li.append(self._PrintPath(i))
				return str(li)
			elif isinstance(data[0], Node):
				return self.prettyAttrNode(alias, data)

			elif isinstance(data[0], Edge):
				return self.prettyAttrEdge(alias, data)
			else:
				return str(data)
		else:
			return str(data)

	def prettyAttr(self, data: DataItem):
		if data.asAttr().type in [ResultType.RESULT_TYPE_PATH, ResultType.RESULT_TYPE_EDGE,
								  ResultType.RESULT_TYPE_NODE]:
			propertList = ["#", data.alias]
			y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
			for i, adata in enumerate(data.asAttr().values):
				ret = self._formatAttrListData(adata, data.alias)
				y.add_row([i + 1, ret])
			print(y)
		else:
			propertList = ["#", data.alias]
			y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
			if len(data.asAttr().values) > 0:
				for i, adata in enumerate(data.asAttr().values):
					ret = self._formatAttrListData(adata, data.alias)
					y.add_row([i + 1, ret])
			else:
				y.add_row([1, str(data.asAttr().values)])
			print(y)

	def prettyArray(self, data: DataItem):
		propertList = [data.alias]
		y = PrettyTable(propertList, title=f"Alias: {data.alias} AliasType: {data.type}", align="l")
		for array in data.asArray().elements:
			y.add_row([json.dumps(array, ensure_ascii=False)])
		print(y)

	def prettyTree(self, plans: ExplainPlan):
		def buildTree(plans: [PlanNode]):
			if plans == None or len(plans) == 0:
				return TreeNode()
			tree = TreeNode(plans[0])
			del (plans[0])
			for i in range(tree.explain.children_num):
				node = buildTree(plans)
				tree.childNodes.append(node)
			return tree

		root = buildTree(plans.planNodes)
		self._printTree(root, root.explain.id, True)
		ret = self.tree.show(stdout=False)
		propertList = ["data"]
		y = PrettyTable(propertList, title=f"Alias: tree AliasType: ExplainPlan", align="l")
		y.add_row([ret])
		print(y)

	def _printTree(self, root: TreeNode, parter, isRoot=False):
		if isRoot:
			self.tree.create_node(root.explain.uql, parter)
		else:
			self.tree.create_node(root.explain.uql, identifier=root.explain.id, parent=parter)
		if len(root.childNodes) > 0:
			parter = root.explain.id
			for child in root.childNodes:
				self._printTree(child, parter)
		else:
			return




from ultipa.structs.HDC import HDCGraph
from ultipa.types.types import HDCBuilder
from ultipa.utils.convert import convertTableToDict,converToHDCGraph
from ultipa.configuration import RequestConfig
from ultipa.utils import  CommandList,UQLMAKER
from ultipa.types import ULTIPA
from ultipa.types.types_response import *
from ultipa.operate.base_extra import BaseExtra

REPLACE_KEYS = {
	"graph": "name",
}
class HDCExtra(BaseExtra):
    '''
    	Processing class that defines settings for HDC related operations.
    '''
    def createHDCGraphBySchema(self,
                               builder:HDCBuilder,
                               config: RequestConfig = RequestConfig())->JobResponse:
        '''
        create HDCGraph

        Args:
            builder: The object of HDCBuilder

            config: An object of RequestConfig class

        Returns:
            JobResponse
        '''

        command=CommandList.createHDCGraph
        uqlmarker=UQLMAKER(command=command,commonParams=config)
        def getschemalist(schemas: []):
            schemalist = []
            for data in schemas:
                for key,value in data.items():
                    if key== '*':
                        schemalist.append(f'"{key}":{value}')
                    else :
                        schemalist.append(f'{key}:{value}')
            return schemalist

        nodeSchemas = ','.join(getschemalist(builder.nodeSchema))
        edgeSchemas = ','.join(getschemalist(builder.edgeSchema))
        commP = '{nodes:{%s},edges:{%s},query:"query",type:"Graph",update:"%s",default:true}' % (
        nodeSchemas, edgeSchemas,builder.syncType)

        uqlmarker.setCommandParams([builder.hdcGraphName,commP])
        uqlmarker.addParam(key='to', value=f'{builder.hdcServerName}')
        res = self.uqlSingle(uqlmarker)
        result = [{'new_job_id': ''}]
        if res.status.code == ErrorCode.SUCCESS and res.items:
            result = convertTableToDict(res.alias(res.aliases[0].alias).entities.rows,
                                        res.alias(res.aliases[0].alias).entities.headers)

        return JobResponse(jobId=result[0].get('new_job_id'), status=res.status, statistics=res.statistics)

    def showHDCGraph(self,config: RequestConfig = RequestConfig())->List[HDCGraph]:

        '''
        show HDCGraph

        Args:

            config: An object of RequestConfig class

        Returns:
            List[HDCGraph]
        '''

        command=CommandList.showHDCGraph
        uqlmaker=UQLMAKER(command=command , commonParams=config)
        uqlmaker.setCommandParams("")
        res = self.uqlSingle(uqlmaker)
        if res.status.code == ULTIPA.ErrorCode.SUCCESS:
            if res.aliases:

                proj = convertTableToDict(res.alias(res.aliases[0].alias).entities.rows,res.alias(res.aliases[0].alias).entities.headers)
                projres = converToHDCGraph(proj)
                return projres
        return res

    def dropHDCGraph(self,hdcGraphName:str,config: RequestConfig = RequestConfig())->Response:

        '''
        drop HDCGraph

        Args:

            hdcGraphName: The name of HDCGraph

            config: An object of RequestConfig class

        Returns:
            Response
        '''

        command=CommandList.dropHDCGraph
        uqlmaker = UQLMAKER(command=command, commonParams=config)
        uqlmaker.setCommandParams(hdcGraphName)
        res = self.uqlSingle(uqlmaker)
        return res




# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 10:34
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : DBType.py
from enum import Enum

from ultipa.proto import ultipa_pb2

class DBType(Enum):
	'''
	    Data class for data type
	'''
	DBNODE = ultipa_pb2.DBNODE
	DBEDGE = ultipa_pb2.DBEDGE
	DBGLOBAL = ultipa_pb2.DBGLOBAL


from abc import abstractmethod

from ultipa import BaseModel


class ExportListener():
    @abstractmethod
    def onReady(self):
        pass

    @abstractmethod
    def next(self):
        pass

    @abstractmethod
    def onError(self):
        pass

    @abstractmethod
    def onComplete(self):
        pass




from ultipa import DBType
from ultipa.utils.convert import convertTableToDict
from ultipa.operate.base_extra import BaseExtra
from ultipa.utils import UQLMAKER, CommandList
from ultipa.types import ULTIPA_REQUEST, ULTIPA
from ultipa.types.types_response import *
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.utils.noneCheck import checkNone

JSONSTRING_KEYS = ["graph_privileges", "system_privileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class TruncateExtra(BaseExtra):
    '''
        Processing class that defines settings for advanced operations on graphset.
    '''

    def truncate(self,
                 params: ULTIPA_REQUEST.TruncateParams,
                 config: RequestConfig = RequestConfig()) -> Response:
        '''
        Truncate graphset.

        Args:
            params: The truncate parameters; the attribute graphName is mandatory, schemaName and dbType are optional while they must be set together.

            config: An object of RequestConfig class.

        Returns:
            Response

        '''

        if checkNone(params.graphName):
            return Response(Status(code=ULTIPA.ErrorCode.PARAM_ERROR,message='The graphName of params cannot be None'))
        else :
            if (checkNone(params.dbType) and  not checkNone(params.schemaName)) or (not checkNone(params.dbType) and checkNone(params.schemaName)):
                return Response(status=ULTIPA.Status(code = ULTIPA.ErrorCode.PARAM_ERROR,
                                                   message="The schemaName and dbType of params must be set together"))
        command = CommandList.truncate
        config.graph = params.graphName
        uqlMaker = UQLMAKER(command, commonParams=config)
        uqlMaker.addParam("graph", params.graphName)

        if params.dbType is not None:
            if params.dbType == DBType.DBNODE:
                if params.schemaName == "*" or params.schemaName is None:
                    uqlMaker.addParam("nodes", "*")
                else:
                    uqlMaker.addParam("nodes", "@" + params.schemaName, notQuotes=True)
            if params.dbType == DBType.DBEDGE:
                if params.schemaName == "*" or params.schemaName is None:
                    uqlMaker.addParam("edges", "*")
                else:
                    uqlMaker.addParam("edges", "@" + params.schemaName, notQuotes=True)

        return self.uqlSingle(uqlMaker)

    def compact(self, graphName: str,
                config: RequestConfig = RequestConfig()) -> JobResponse:
        '''
        Compact graphshet.

        Args:
            graphName: The name of graphset

            config: An object of RequestConfig class

        Returns:
            JobResponse

        '''
        command = CommandList.compact
        uqlMaker = UQLMAKER(command, commonParams=config)
        uqlMaker.addParam("graph", graphName)
        result = self.uqlSingle(uqlMaker)
        jobResponse = JobResponse()
        jobResponse.statistics = result.statistics
        jobResponse.status = result.status
        if result.items:
            res = convertTableToDict(result.alias(result.aliases[0].alias).entities.rows, result.alias(result.aliases[0].alias).entities.headers)

            jobResponse.jobId = res[0]['new_job_id']
            return jobResponse
        else:
            return jobResponse



import time
from typing import List

from ultipa.utils.checkStrTime import is_valid_date


class FilterOpt:
    EQ = "$eq"
    LT = "$lt"
    LTE = "$lte"
    GT = "$gt"
    GTE = "$gte"
    BT = "$bt"
    IN = "$in"
    OR = "$or"
    AND = "$and"
    NIN = "$nin"

class FilterBase:
    pass



class Filter(FilterBase):
    def __init__(self, name: str, value: object=None):
        self.name = name
        self.value = value

    def builder(self):
        if self.name == 'id':
            self.name = '_id'
        try:
            if is_valid_date(self.value):
                filter = '{%s:{%s:"%s"}}' % (self.name, self.opt, self.value)
            else:
                filter = '{%s:{%s:%s}}' % (self.name, self.opt, self.value)
            return filter
        except Exception as e:
            print(e)
            return "%s"%self.value


class EqFilter(Filter):
    def __init__(self, name: str, value: any):
        super().__init__(name,value)
        self.opt = FilterOpt.EQ

    def builder(self):
        if self.name == 'id':
            self.name = '_id'
        try:
            filter = '{%s:{%s:"%s"}}' % (self.name, self.opt, self.value)
            return filter
        except:
            return "%s"%self.value


class LtFilter(Filter):
    def __init__(self, name: str, value: any):
        super().__init__(name, value)
        self.opt = FilterOpt.LT

class LteFilter(Filter):
    def __init__(self, name: str, value: any):
        super().__init__(name, value)
        self.opt = FilterOpt.LTE

class GtFilter(Filter):
    def __init__(self, name: str, value: any):
        super().__init__(name, value)
        self.opt = FilterOpt.GT

class GteFilter(Filter):
    def __init__(self, name: str, value: any):
        super().__init__(name, value)
        self.opt = FilterOpt.GTE

class BtFilter(Filter):
    def __init__(self, name: str, value: object):
        super().__init__(name, value)
        self.opt = FilterOpt.BT

    def builder(self):
        if self.name == 'id':
            self.name = '_id'
        try:
            filter = '{%s:{%s:%s}}' % (self.name, self.opt, self.value)
            return filter
        except:
            return "%s"%self.value


class InFilter(Filter):
    def __init__(self, name: str, value: object):
        super().__init__(name, value)
        self.opt = FilterOpt.IN

    def builder(self):
        if self.name == 'id':
            self.name = '_id'
        try:
            filter = '{%s:{%s:%s}}' % (self.name, self.opt, self.value)
            return filter
        except:
            return "%s"%self.value


class NinFilter(Filter):
    def __init__(self, name: str, value: object):
        super().__init__(name, value)
        self.opt = FilterOpt.NIN

    def builder(self):
        if self.name == 'id':
            self.name = '_id'
        try:
            filter = '{%s:{%s:%s}}' % (self.name, self.opt, self.value)
            return filter
        except:
            return "%s"%self.value


class OrFilter(FilterBase):
    def __init__(self,value: List[Filter]):
        # super().__init__(name)
        self.value=value
        self.opt = FilterOpt.OR

    def builder(self):
        # if self.name == 'id':
        #     self.name = '_id'
        filter = '{%s:[%s]}' % (self.opt,','.join([f.builder() for f in self.value]))
        return filter

class AndFilter(FilterBase):
    def __init__(self, value: List[Filter]):
        # super().__init__(name)
        self.value=value
        self.opt = FilterOpt.AND

    def builder(self):
        # if self.name == 'id':
        #     self.name = '_id'
        filter = '{%s:[%s]}' % (self.opt,','.join([f.builder() for f in self.value]))
        # filter = filter.replace('"',"")
        return filter


# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Privilege.py
from enum import Enum, auto
from typing import List
from ultipa.structs.BaseModel import BaseModel


class PrivilegeLevel(Enum):
    SYSTEM = auto()
    GRAPH = auto()


class Privilege(BaseModel):
    '''
        Data class for Privilege.
    '''

    def __init__(self, name: str = None, level: PrivilegeLevel = None):
        self.name = name
        self.level = level



import datetime
import math

from ultipa.proto.ultipa_pb2 import ListData, SetData, AttrListData
from ultipa.structs import Property
from ultipa.types import ULTIPA
from struct import *

from ultipa.utils import nullValue
from ultipa.utils.common import DateTimeIs2Str, Float_Accuracy, TimestampIs2Str
from ultipa.utils.ultipa_datetime import UltipaDatetime
from ultipa.utils.errors import ParameterException, ServerException, SerializeException, checkError
import ast

class _Serialize:
	'''
	Configuration class that defines settings for serialization.
	'''
	def __init__(self, type, value, name=None, export=False, timeZoneOffset=None, subTypes=None, timeZone=None):
		self.type = type
		self.value = value
		self.subTypes = subTypes
		self.name = name
		self.export = export
		self.timeZoneOffset = timeZoneOffset
		self.timeZone = timeZone

	def serialize(self):
		if self.value is None and not self.subTypes:
			self.value = nullValue.nullValue(self.type)
			return self.value

		# if self.value is None:
		# 	self.value = nullValue.nullValue(self.type)
		# 	return self.value
		if self.type == ULTIPA.UltipaPropertyType.BLOB.value:
			return str(self.value).encode()

		if self.type == ULTIPA.UltipaPropertyType.STRING.value or self.type == ULTIPA.UltipaPropertyType.TEXT.value or self.type == ULTIPA.UltipaPropertyType.POINT.value or self.type == ULTIPA.UltipaPropertyType.DECIMAL.value:
			if isinstance(self.value, str):
				return self.value.encode()
			elif isinstance(self.value, bytes):
				return self.value
			else:
				return str(self.value).encode()

		elif self.type == ULTIPA.UltipaPropertyType.BOOL.value:
			return str(int(self.value)).encode()

		elif self.type == ULTIPA.UltipaPropertyType.INT32.value:
			if self.value == '':
				self.value = 0
			try:
				upret = pack('>i', int(self.value))
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")

		elif self.type == ULTIPA.UltipaPropertyType.UINT32.value:
			if self.value == '':
				self.value = 0
			try:
				upret = pack('>I', int(self.value))
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")

		elif self.type == ULTIPA.UltipaPropertyType.INT64.value:
			if self.value == '':
				self.value = 0
			try:
				upret = pack('>q', int(self.value))
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")


		elif self.type == ULTIPA.UltipaPropertyType.UINT64.value:
			if self.value == '':
				self.value = 0
			try:
				upret = pack('>Q', int(self.value))
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")

		elif self.type == ULTIPA.UltipaPropertyType.FLOAT.value:
			if self.value == '':
				self.value = 0

			try:
				upret = pack('>f', self.value)
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")

		elif self.type == ULTIPA.UltipaPropertyType.DOUBLE.value:
			if self.value == '':
				self.value = 0
			try:
				upret = pack('>d', self.value)
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")

		elif self.type == ULTIPA.UltipaPropertyType.DATETIME.value:
			if self.value is None:
				self.value = 0
			else:
				self.value = UltipaDatetime.datetimeStr2datetimeInt(self.value)
			try:
				upret = pack('>Q', self.value)
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")
			return upret

		elif self.type == ULTIPA.UltipaPropertyType.TIMESTAMP.value:
			if self.value is None:
				self.value = 0
			else:
				if isinstance(self.value, datetime.datetime):
					self.value = int(self.value.timestamp())
				if isinstance(self.value, str):
					self.value = UltipaDatetime.timestampStr2timestampInt(self.value, self.timeZone,
																		  self.timeZoneOffset)
			try:
				upret = pack('>I', self.value)
				return upret
			except Exception as e:
				error = checkError(e.args[0])
				raise SerializeException(err=f"property [%s],value=%s {error}")

		elif (self.type == ULTIPA.UltipaPropertyType.LIST.value or self.type == ULTIPA.UltipaPropertyType.SET.value)and self.subTypes != None and len(self.subTypes) > 0:
			listData = ListData()
			if self.value == None:
				listData.is_null = True
				self.value = []
			for i, v in enumerate(self.value):
				if isinstance(self.subTypes[0],str):
					type = Property._getPropertyTypeByString(self.subTypes[0]).value
				else:
					type = self.subTypes[0]
				listData.values.append(_Serialize(type, v, timeZone=self.timeZone,
										  timeZoneOffset=self.timeZoneOffset).serialize())
			return listData.SerializeToString()


		elif self.type == ULTIPA.UltipaPropertyType.MAP.value and self.subTypes != None and len(self.subTypes) > 0:
			setData = SetData()
			if self.value == None:
				setData.is_null = True
				self.value = set()
			for i, v in enumerate(self.value):
				setData.values.add(_Serialize(self.subTypes[0], v).serialize())
			return setData.SerializeToString()

	def unserialize(self):
		try:
			if self.type == ULTIPA.UltipaPropertyType.NULL.value:
				return None

			if nullValue.isNullValue(self.value, self.type):
				return None

			elif self.type == ULTIPA.UltipaPropertyType.BLOB.value:
				return list(bytearray(self.value))

			elif self.type == ULTIPA.UltipaPropertyType.STRING.value or self.type == ULTIPA.UltipaPropertyType.DECIMAL.value or self.type == ULTIPA.UltipaPropertyType.TEXT.value or type is None:
				return self.value.decode()

			elif self.type == ULTIPA.UltipaPropertyType.BOOL.value:
				return self.value.decode()=="1"

			elif self.type == ULTIPA.UltipaPropertyType.POINT.value:
				if len(self.value) != 16:
					raise ParameterException("deserialize point type error: length != 16")
				latitude_v = self.value[:8]
				latitude = self.unpackDouble(latitude_v)

				longitude_v = self.value[8:16]
				longitude = self.unpackDouble(longitude_v)
				if math.isnan(longitude) or math.isnan(latitude):
					return None
				return f"POINT({latitude} {longitude})"



			elif self.type == ULTIPA.UltipaPropertyType.INT32.value:
				if len(self.value) >= 4:
					ls = len(self.value) // 4 * 'i' or 'i'
				elif len(self.value) == 2:
					ls = len(self.value) // 2 * 'h' or 'h'
				else:
					ls = 'h'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.UINT32.value:
				ls = len(self.value) // 4 * 'I' or 'I'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.INT64.value:
				ls = len(self.value) // 8 * 'q' or 'q'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.UINT64.value:
				ls = len(self.value) // 8 * 'Q' or 'Q'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.FLOAT.value:
				ls = len(self.value) // 4 * 'f' or 'f'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				if Float_Accuracy:
					return round(ret, 7)
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.DOUBLE.value:
				ls = len(self.value) // 8 * 'd' or 'd'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.DATETIME.value:
				ls = len(self.value) // 8 * 'Q' or 'Q'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				if DateTimeIs2Str:
					ret = UltipaDatetime.datetimeInt2datetimeStr(ret)
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.TIMESTAMP.value:
				ls = len(self.value) // 4 * 'I' or 'I'
				upret = unpack(f'>{ls}', self.value)
				ret = upret[0]
				if TimestampIs2Str:
					ret = UltipaDatetime.timestampInt2timestampStr(ret, self.timeZone, self.timeZoneOffset)
				if nullValue.isNullValue(self.value, self.type):
					return None
				return ret
			elif self.type == ULTIPA.UltipaPropertyType.LIST.value and self.subTypes != None and len(self.subTypes) > 0:
				ret = []
				listData = ListData()
				listData.ParseFromString(self.value)
				if listData.is_null == True:
					return None
				for v in listData.values:
					ret.append(_Serialize(self.subTypes[0], v, timeZone=self.timeZone,
										  timeZoneOffset=self.timeZoneOffset).unserialize())
				return ret

			elif self.type == ULTIPA.UltipaPropertyType.SET.value:
				ret = set()
				setData = SetData()
				setData.ParseFromString(self.value)
				if setData.is_null == True:
					return None
				for value in setData.values:
					ret.add(_Serialize(self.subTypes[0], value, timeZone=self.timeZone,
										  timeZoneOffset=self.timeZoneOffset).unserialize())
				return list(ret)
				# ret = ast.literal_eval(self.value.decode("utf-8"))
				# return ret
			else:
				raise ServerException('Server returned type error')

		except Exception as e:
			raise ParameterException(err=e)

	def packDouble(self,value):
		return pack('>d', value)

	def unpackDouble(self,value):
		ls = len(value) // 8 * 'd' or 'd'
		upret = unpack(f'>{ls}', value)
		latitude = upret[0]
		return latitude


	def setDefaultValue(self):
		if self.type == ULTIPA.UltipaPropertyType.STRING.value:
			self.value = ""
			return
		elif self.type == ULTIPA.UltipaPropertyType.INT32.value:
			self.value = 0
			return
		elif self.type == ULTIPA.UltipaPropertyType.UINT32.value:
			self.value = 0
			return
		elif self.type == ULTIPA.UltipaPropertyType.INT64.value:
			self.value = 0
			return
		elif self.type == ULTIPA.UltipaPropertyType.UINT64.value:
			self.value = 0
			return
		elif self.type == ULTIPA.UltipaPropertyType.FLOAT.value:
			self.value = 0
			return
		elif self.type == ULTIPA.UltipaPropertyType.DOUBLE.value:
			self.value = 0
			return
		elif self.type == ULTIPA.UltipaPropertyType.TEXT.value:
			self.value = ""
			return
		elif self.type == ULTIPA.UltipaPropertyType.DATETIME.value:
			self.value = "1970-01-01"
			return
		elif self.type == ULTIPA.UltipaPropertyType.TIMESTAMP.value:
			self.value = "1970-01-01"
			return



from ultipa import DBType, UltipaPropertyType
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types.types_response import *
from ultipa.types.types import Status, ErrorCode
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.noneCheck import checkNone
from ultipa.utils.propertyUtils import getPropertyTypesDesc
from typing import Tuple, List, Dict

BOOL_KEYS = ["index", "lte"]
REPLACE_KEYS = {
    "name": "propertyName",
    "type": "propertyType",
}


class PropertyExtra(BaseExtra):
    '''
    Processing class that defines settings for property related operations.
    '''

    def showProperty(self, dbType: DBType = None, schemaName: str = None,
                     config: RequestConfig = RequestConfig()) -> Dict[str, List[Property]]:
        '''
        Show all Node or Edge Schema properties.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            schemaName: The name of schema

            config: An object of RequestConfig class

        Returns:
            Dict[str, List[Property]]

        '''
        if not checkNone(schemaName) and (dbType == None or dbType == DBType.DBGLOBAL):
            return Response(status=Status(code=ErrorCode.PARAM_ERROR, message='When schemaName is not None , dbType cannot be None and DBGLOBAL'))
        if dbType == None or dbType == DBType.DBGLOBAL:

            command = CommandList.showProperty
            commandp = ''
        else:
            if dbType == DBType.DBNODE:
                command = CommandList.showNodeProperty
            elif dbType == DBType.DBEDGE:
                command = CommandList.showEdgeProperty
            else:
                raise TypeError("dbType must be an instance of DBType.DBNODE or DBType.DBEDGE or DBType.DBGLOBAL")
            if schemaName:
                commandp = ['@' + f"`{schemaName}`"]
            else:
                commandp = ''

        uqlMaker = UQLMAKER(command=command, commonParams=config)
        uqlMaker.setCommandParams(commandP=commandp)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        propertyDict = {}
        nodeProperty = res.alias('_nodeProperty').asProperties()
        edgeProeprty = res.alias('_edgeProperty').asProperties()
        if nodeProperty :
            propertyDict['nodePropertys'] = nodeProperty
        if edgeProeprty :
            propertyDict['edgePropertys'] = edgeProeprty

        resprop = propertyDict['nodePropertys'] if dbType == DBType.DBNODE and 'nodePropertys' in propertyDict \
            else  propertyDict['edgePropertys'] if dbType == DBType.DBEDGE and 'edgePropertys' in propertyDict else propertyDict

        return resprop

    def showNodeProperty(self, schemaName: str = None,
                         config: RequestConfig = RequestConfig()) -> List[Property]:
        '''
        Show all Node Schema properties.

        Args:
            schemaName: The name of schema

            config: An object of RequestConfig class

        Returns:
            List[Property]

        '''
        res = self.showProperty(schemaName=schemaName, dbType=DBType.DBNODE, config=config)
        res = res if res else None
        return res

    def showEdgeProperty(self, schemaName: str = None,
                         config: RequestConfig = RequestConfig()) -> List[Property]:
        '''
        Show all Edge Schema properties.

        Args:
            schemaName: The name of schema

            config: An object of RequestConfig class

        Returns:
            List[Property]

        '''

        res = self.showProperty(schemaName=schemaName, dbType=DBType.DBEDGE, config=config)
        res = res if res else None
        return res

    def createProperty(self, dbType: DBType, schemaName: str, property: Property,
                       config: RequestConfig = RequestConfig()) -> Response:
        '''
        Create a property.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            schemaName: The name of schema

            property:   The property to be created; the attributes name and type (and subType if the type is SET or LIST) are mandatory, encrypt and description are optional.

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        if not isinstance(property,Property):
            message = 'prop cannot None' if property is None else 'prop type exception'
            return Response(status=Status(code=ErrorCode.FAILED, message=message))
        if (property.type == UltipaPropertyType.LIST or property.type == UltipaPropertyType.SET) and checkNone(
                property.subType):
            return Response(status=Status(code=ErrorCode.PARAM_ERROR, message='if the type is SET or LIST,The subType of property cannot be None'))
        if checkNone(schemaName):
            return Response(status=Status(code=ErrorCode.PARAM_ERROR, message='The schemaName cannot be None'))
        if dbType == DBType.DBGLOBAL or checkNone(dbType):
            message = dbType == DBType.DBGLOBAL and  'The dbType cannot be DBType.DBGLOBAL' or 'The dbType cannot be None'
            return Response(status=Status(code=ErrorCode.PARAM_ERROR, message=message))
        command = dbType == DBType.DBNODE and CommandList.createNodeProperty or CommandList.createEdgeProperty
        property.type = property.getStringByPropertyType(property.type)
        if property.subType and isinstance(property.subType,list):
            property.subType = [property.getStringByPropertyType(property.subType[0])]
        commandP = ["@" + f"`{schemaName}`", f"`{property.name}`",
                    getPropertyTypesDesc(property.type, property.subType)]

        if property.description:
            commandP.append(property.description)

        uqlMaker = UQLMAKER(command=command, commonParams=config)
        uqlMaker.setCommandParams(commandP=commandP)
        if property.encrypt is not None:
            uqlMaker.addParam('encrypt',property.encrypt,required=False )
        res = self.uqlSingle(uqlMaker)
        return res

    def dropProperty(self, dbType: DBType, property:Property,
                     config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop a property.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            property: An object of Property class

            config: An object of config class

        Returns:
            Response

        '''
        if checkNone(property.schema) :
            return Response(status=Status(code=ErrorCode.PARAM_ERROR,message='The shcema of property cannot be None'))
        schemaName = '*' if property.schema == '' or property.schema is None else property.schema
        command = dbType == DBType.DBNODE and CommandList.dropNodeProperty or CommandList.dropEdgeProperty
        commandP = "@`%s`.`%s`" % (schemaName, property.name)
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        uqlMaker.setCommandParams(commandP=commandP)
        res = self.uqlSingle(uqlMaker)
        return res

    def alterProperty(self, dbType: DBType, originProp: Property, newProp: Property,
                      config: RequestConfig = RequestConfig()) -> Response:
        '''
        Alter a property.

        Args:
            dbType: The DBType of data (DBNODE or DBEDGE)

            originProp: The property to be altered; the attributes name and schema (writes * to specify all schemas) are mandatory.

            newProp: A Property object used to set the new name and/or description for the property.

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        if checkNone(originProp.name) or checkNone(originProp.schema):
            return Response(status=Status(code= ErrorCode.PARAM_ERROR,message='The name and schema of originProp cannot be None'))
        command = dbType == DBType.DBNODE and CommandList.alterNodeProperty or CommandList.alterEdgeProperty
        commandP = "@`%s`.`%s`" % (originProp.schema, originProp.name)
        update_dict = {}
        if newProp.name:
            update_dict.setdefault('name', newProp.name)
        if newProp.description:
            update_dict.update({'description': newProp.description})
        uqlMaker = UQLMAKER(command=command, commonParams=config)
        uqlMaker.setCommandParams(commandP)
        uqlMaker.addParam("set", update_dict)
        res = self.uqlSingle(uqlMaker)
        return res




import time

import json
from typing import Tuple, List, Dict

from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types.types_response import *
from ultipa.types import ULTIPA
from ultipa.types.types import Status, ErrorCode
from ultipa.types.types_response import Response
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertToGraph
from ultipa.structs.GraphSet import GraphSet
from ultipa.utils.noneCheck import checkNone

REPLACE_KEYS = {
	"graph": "name",
}


class GraphExtra(BaseExtra):
	'''
	Processing class that defines settings for GraphSet related operations.
	'''

	def showGraph(self, config: RequestConfig = RequestConfig()) ->List[GraphSet]:
		'''
		Args:
			config: An object of RequestConfig class

		Returns:
			List[GraphSet]
		'''

		uqlMaker = UQLMAKER(command=CommandList.showGraphMore, commonParams=config)
		uqlMaker.setCommandParams("")
		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS))
		if res.status.code == ULTIPA.ErrorCode.SUCCESS :
			res.data = convertToGraph(res)
		return res.data

	def getGraph(self, graphName: str,
				 config: RequestConfig = RequestConfig()) -> GraphSet:
		'''
		Args:
			graphName: The name of GraphSet

			config: An object of RequestConfig class

		Returns:
			GraphSet
		'''

		uqlMaker = UQLMAKER(command=CommandList.showGraph, commonParams=config)
		uqlMaker.setCommandParams(graphName)
		uqlMaker.addParam(key='more',value=graphName)
		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS))
		if res.status.code == ULTIPA.ErrorCode.SUCCESS:
			res.data = convertToGraph(res)
		else:
			res.data = None
		return res.data

	def createGraph(self, graphSet: GraphSet,
					config: RequestConfig = RequestConfig()) -> Response:
		'''
		Create a GraphSet.

		Args:
			graphSet: The graph to be created; the attribute name is mandatory, shards, partitionBy and description are optional.

			config: An object of RequestConfig class

		Returns:
			Response
		'''
		if checkNone(graphSet.name):
			return Response(status=Status(code = ErrorCode.PARAM_ERROR,message='The name of graphSet cannot be None'))
		shardslist = f" shards {[int(shard) if shard.isdigit() else 0 for shard in graphSet.shards]}" if isinstance(graphSet.shards,list) and len(graphSet.shards) >0 else ''
		partitionby = f" PARTITION BY HASH({graphSet.partitionBy})" if graphSet.partitionBy is not None else ''
		comment = f" COMMENT '{graphSet.description}'" if graphSet.description else ''
		gqlMaker = CommandList.createGraphGql + f"{graphSet.name} {{ }} {partitionby} {shardslist} {comment}"
		res = self.gql(gqlMaker,config = config)
		return res


	def dropGraph(self, graphName: str,
				  config: RequestConfig = RequestConfig()) -> Response:
		'''
		Drop a GraphSet.

		Args:
			graphName: The name of GraphSet

			config: An object of RequestConfig class

		Returns:
			Response
		'''

		uqlMaker = UQLMAKER(command=CommandList.dropGraph, commonParams=config)
		uqlMaker.setCommandParams(graphName)
		res = self.uqlSingle(uqlMaker)
		return res

	def alterGraph(self, graphName: str, alterGraphSet: GraphSet,
			   config: RequestConfig = RequestConfig()) -> Response:
		'''
		Args:
			graphName: The orignal name of GraphSet

			alterGraphSet: A GraphSet object used to set the new name and/or description for the graph. the attribute name is mandatory

			config: An object of RequestConfig class

		Returns:
			Response
		'''
		if checkNone(alterGraphSet.name):
			return Response(status=Status(code=ErrorCode.PARAM_ERROR,message='The name of alterGraphSet cannot be None'))
		config.graph = graphName
		uqlMaker = UQLMAKER(command=CommandList.alterGraph, commonParams=config)
		uqlMaker.setCommandParams(graphName)
		data = {"name": alterGraphSet.name}
		if alterGraphSet.description is not None:
			data.update({'description': alterGraphSet.description})
		uqlMaker.addParam("set", data)
		res = self.uqlSingle(uqlMaker)
		return res

	def hasGraph(self, graphName: str, config: RequestConfig = RequestConfig()) -> bool:
		'''
		Check if graph exists or not.

		Args:
			graphName: The name of GraphSet

			config: An object of RequestConfig class

		Returns:
			bool
		'''

		graphsdata = self.showGraph(config)
		for graphs in graphsdata:
			if (graphs.name == graphName):
				return True

		return False

	def createGraphIfNotExist(self, graphSet: GraphSet,
							  config: RequestConfig = RequestConfig()) -> ResponseWithExistCheck:
		'''
		Checks if graph exists or not, if graph does not exist then creates new.

		Args:
			graphSet: The object of graphSet

			config: An object of RequestConfig class

		Returns:
			ResponseWithExistCheck
		'''

		if (self.hasGraph(graphSet.name, config)) == True:
			return ResponseWithExistCheck(exist=True,response=Response())

		else:
			res = self.createGraph(graphSet,config = config)
			return ResponseWithExistCheck(exist=False,response=res)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:20
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Path.py
from typing import Dict, List

from ultipa.structs.Schema import Schema
from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.Edge import Edge
from ultipa.structs.Node import Node


class Path(BaseModel):
	'''
	    Data class for path.
	'''
	def __init__(self,nodeUuids: List[int]=None,
				 edgeUuids: List[int]=None,
				 nodes: Dict[int, Node] = {},
				 edges: Dict[int, Edge] = {}
				):
		self.nodeUuids = nodeUuids
		self.edgeUuids = edgeUuids
		self.nodes = nodes
		self.edges = edges

	def length(self):
		return len(self.edgeUuids)

class PathAlias:
	'''
	    Data class for path with alias.
	'''
	def __init__(self, alias: str, paths: List[Path] = None):
		self.alias = alias
		if paths is None:
			paths = []
		self.paths = paths

	def length(self):
		return len(self.paths)

	def getNodes(self):
		nodes = [i.nodeUuids for i in self.paths]
		return nodes

	def getEdges(self):
		edges = [i.edgeUuids for i in self.paths]
		return edges


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Graph.py
from ultipa.structs.BaseModel import BaseModel
class Replica(BaseModel):
    addr: str
    status: str
    streamAddr: str
    lastHeartBeat: str

    def __init__(self, addr: str = '', status: str = '', streamAddr: str = '',
                 lastHeartBeat: str = ''):
        self.addr = addr
        self.status = status
        self.streamAddr = streamAddr
        self.lastHeartBeat = lastHeartBeat







from ultipa.structs.BaseModel import BaseModel


class HDCGraph(BaseModel):
    '''
        Data class for HDCGraph.
    '''

    def __init__(self,
                 name: str = None,
                 graphName: str = None,
                 status: str= None,
                 stats: str= None,
                 isDefault: str= None,
                 hdcServerName: str= None,
                 hdcServerStatus: str= None,
                 config: str= None):
        self.name = name
        self.isDefault = isDefault
        self.graphName = graphName
        self.status = status
        self.stats = stats
        self.hdcServerName = hdcServerName
        self.hdcServerStatus = hdcServerStatus
        self.config = config


class HDCSyncType(BaseModel):
    SYNC = "sync"
    ASYNC = "async"
    SCTATIC = "static"


# -*- coding: utf-8 -*-
# @Time    : 2024/05/17 10:56
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Top.py
from ultipa.structs.BaseModel import BaseModel


class Top(BaseModel):
	'''
	    Data class for Top Processes.
	'''

	def __init__(self, process_id: str, process_query: str,duration: str,status:str):
		self.process_id=process_id
		self.process_query=process_query
		self.duration=duration
		self.status=status


# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 09:20
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : connectionBase.py

import logging
import time
import schedule

from ultipa import ParameterException
from ultipa.configuration.UltipaConfig import UltipaConfig
from ultipa.connection.clientInfo import ClientInfo
from ultipa.connection.clientType import ClientType
from ultipa.connection.hostManagerControl import HostManagerControl, RAFT_GLOBAL
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.proto import ultipa_pb2
from ultipa.utils import CommandList
from ultipa.utils.ultipaSchedule import run_continuously


class ConnectionBase:
	'''
		A base class that defines settings for an Ultipa connection.

	'''

	def __init__(self, host: str, defaultConfig: UltipaConfig, crt: str = None):
		self.host = host
		self.username = defaultConfig.username
		self.password = defaultConfig.password
		self.crt = crt
		self.defaultConfig = defaultConfig
		self.runSchedule: object = None
		self.crt = None
		if crt:
			try:
				with open(f'{crt}', 'rb') as f:
					self.crt = f.read()
			except Exception as e:
				raise ParameterException(err=e)
		self.hostManagerControl = HostManagerControl(self.host, self.username, self.password,
													 self.defaultConfig.maxRecvSize, self.crt,
													 )

		self.defaultConfig.defaultGraph = defaultConfig.defaultGraph or "default"
		self.defaultConfig.timeout = defaultConfig.timeout or 15
		self.graphSetName = self.defaultConfig.defaultGraph
		self.count = 0


	def getGraphSetName(self, currentGraphName: str, uql: str = "", isGlobal: bool = False):
		if isGlobal:
			return RAFT_GLOBAL
		if uql:
			parse = UQLHelper(uql)
			if parse.uqlIsGlobal():
				return RAFT_GLOBAL
			c1 = parse.parseRet.getFirstCommands()
			c2 = f"{c1}().{parse.parseRet.getSecondCommands()}"
			if c2 in [CommandList.mount, CommandList.unmount, CommandList.truncate]:
				graphName = parse.parseRet.getCommandsParam(1)
				if graphName:
					return graphName
		return currentGraphName or self.defaultConfig.defaultGraph

	def getTimeout(self, timeout: int):
		return timeout or self.defaultConfig.timeout

	def getClientInfo(self, clientType: int = ClientType.Default, graphSetName: str = '', uql: str = '',
					  isGlobal: bool = False, ignoreRaft: bool = False, useHost: str = None,
					  timezone=None, timeZoneOffset=None):
		goGraphName = self.getGraphSetName(currentGraphName=graphSetName, uql=uql, isGlobal=isGlobal)
		if not ignoreRaft and not self.hostManagerControl.getHostManger(goGraphName).raftReady:
			self.hostManagerControl.getHostManger(goGraphName).raftReady = False

		clientInfo = self.hostManagerControl.chooseClientInfo(type=clientType, uql=uql, graphSetName=goGraphName,
															  useHost=useHost)
		metadata = clientInfo.getMetadata(goGraphName, timezone, timeZoneOffset)
		return ClientInfo(Rpcsclient=clientInfo.Rpcsclient, Controlsclient=clientInfo.Controlsclient, metadata=metadata,
						  graphSetName=goGraphName, host=clientInfo.host)


	def stopConnectionAlive(self):
		if self.runSchedule != None:
			self.runSchedule.set()

	def keepConnectionAlive(self, timeIntervalSeconds: int = None):
		timeIntervalSeconds = self.defaultConfig.heartBeat if timeIntervalSeconds == None else timeIntervalSeconds

		def test_allconn():
			goGraphName = self.defaultConfig.defaultGraph
			for host in self.hostManagerControl.getAllClientInfos(goGraphName):
				res = host.Controlsclient.SayHello(ultipa_pb2.HelloUltipaRequest(name="test"),
												   metadata=host.getMetadata(goGraphName, None, None))
				# print(host.host,res.message)

		schedule.every().second.do(test_allconn)
		self.runSchedule = run_continuously(timeIntervalSeconds)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:43
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : __init__.py.py

from ultipa.configuration.UltipaConfig import UltipaConfig
from ultipa.configuration.InsertRequestConfig import InsertRequestConfig
from ultipa.configuration.RequestConfig import RequestConfig



# -*- coding: utf-8 -*-
# @Time    : 2023/1/17 9:45 AM
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : null.py
import sys
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils.errors import ParameterException, ServerException, SerializeException, checkError


StringNull = bytes([0x00])
Int32Null = bytes([0x7f,0xff,0xff,0xff])
Uint32Null =bytes([0xff,0xff,0xff,0xff])
Int64Null = bytes([0x7f,0xff,0xff,0xff,0xff,0xff,0xff,0xff])
Uint64Null = bytes([0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff])
FloatNull = bytes([0xff,0xff,0xff,0xff])
DoubleNull = bytes([0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff])
PointNull = bytes([0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff])
BoolNull = bytes([0x2])

def nullValue(type:ULTIPA.UltipaPropertyType):
	'''
	Returns the null value corresponding to different data types

	Args:
		type: The data type

	Returns:

	'''
	if type in [ULTIPA.UltipaPropertyType.STRING.value,ULTIPA.UltipaPropertyType.TEXT.value,ULTIPA.UltipaPropertyType.BLOB.value,ULTIPA.UltipaPropertyType.DECIMAL.value]:
		return StringNull
	elif type == ULTIPA.UltipaPropertyType.INT32.value:
		return Int32Null
	elif type == ULTIPA.UltipaPropertyType.UINT32.value:
		return Uint32Null
	elif type == ULTIPA.UltipaPropertyType.INT64.value:
		return Int64Null
	elif type == ULTIPA.UltipaPropertyType.UINT64.value:
		return Uint64Null
	elif type == ULTIPA.UltipaPropertyType.FLOAT.value:
		return FloatNull
	elif type == ULTIPA.UltipaPropertyType.DOUBLE.value:
		return DoubleNull
	elif type == ULTIPA.UltipaPropertyType.DATETIME.value:
		return Uint64Null
	elif type == ULTIPA.UltipaPropertyType.TIMESTAMP.value:
		return Uint32Null
	elif type == ULTIPA.UltipaPropertyType.POINT.value:
		return PointNull
	elif type == ULTIPA.UltipaPropertyType.BOOL.value:
		return BoolNull
	elif type in [ULTIPA.UltipaPropertyType.LIST.value,ULTIPA.UltipaPropertyType.SET.value]:
		return None
	raise SerializeException(f"not support [{ULTIPA.Property._PropertyReverseMap.get(type)}]")

def isNullValue(v: any, type:ULTIPA.UltipaPropertyType):
	'''
	Judge whether a value is null

	Args:
		v: The value to be judged
		type: The property type of v

	Returns:
		bool

	'''
	try:
		nullV = nullValue(type)
		return nullV == v
	except Exception as e:
		raise SerializeException(e)
	return False






if __name__ == '__main__':
	print(sys.float_info.max)
	print(sys.int_info)
	print("Int32Null：",Int32Null)
	print("Ret：",0X7FFFFFFF==Int32Null)
	print("Uint32Null：",Uint32Null)
	print("Int64Null：",Int64Null)
	print("Uint64Null：",Uint64Null)
	print("FloatNull：",FloatNull)
	print("DoubleNull：",DoubleNull)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/10 18:25
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : UltipaConfig.py
from typing import List
from ultipa.utils.logger import LoggerConfig
from enum import Enum
from ultipa.utils.password2md5 import passwrod2md5

class UltipaConfig:
	'''
    Configuration class used to instantiate Ultipa connection.

    This class stores settings for establishing a connection to an Ultipa server. The change of an UltipaConfig object will update the connection it established.
    
    Args:
        - hosts (List[str]): Required. A comma-separated list of database server IPs or URLs. If a URL does not start with https:// or http://, http:// is prefixed by default.
        - username (str): Required. Username of the host authentication.
        - password (str): Required. Password of the host authentication.
		- passwordEncrypt(str): Password encryption method of the driver. Supports MD5, LDAP and NOTHING.
        - crt (str): The file path of the SSL certificate used for secure connections.
        - defaultGraph (str): Name of the graph to use by default in the database.
        - timeout (int): Connection timeout threshold (in second).
        - heartBeat (int): The heartbeat interval (in millisecond), used to keep the connection alive. Set to 0 to disable.
        - maxRecvSize (int): The maximum size (in MB) of the received data.
        
    '''

	def __init__(self, hosts:List[str]=None,
				 username:str=None,
				 password:str=None,
				 passwordEncrypt:str='MD5',
				 crt:str=None,
				 defaultGraph: str = None,
				 timeout: int = 15,
				 heartBeat: int = 0,
				 maxRecvSize: int = 32):
		if hosts is None:
			hosts = []
		self.hosts = hosts
		self.username = username
		self.defaultGraph = defaultGraph
		self.crt = crt
		self._password=None
		self.passwordEncrypt=passwordEncrypt
		self.timeout = timeout
		self.heartBeat = heartBeat
		self.maxRecvSize = maxRecvSize * 1024 * 1024
		self.password = password
	@property
	def password(self):
		return self._password

	@password.setter
	def password(self, value):
		if self.passwordEncrypt == 'MD5' and value is not None:
			self._password = passwrod2md5(value)
		else:
			self._password = value

	def setDefaultGraphName(self, graph: str):
			self.defaultGraph = graph


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Graph.py
from typing import List

from ultipa.structs.BaseModel import BaseModel


class GraphSet(BaseModel):
    '''
            Data class for graphSet.
    '''

    def __init__(self,
                 name: str = None,
                 id: str = None,
                 totalNodes: int = None,
                 totalEdges: int = None,
                 partitionBy: str = None,
                 status: str = None,
                 shards: List[str] = [],
                 description: str = None,
                 slotNum: int = 0,
                 ):
        self.id = id
        self.name = name
        self.totalNodes = totalNodes
        self.totalEdges = totalEdges
        self.shards = shards
        self.partitionBy = partitionBy
        self.status = status
        self.description = description
        self.slotNum = slotNum



# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : User.py
import json
from datetime import  datetime
from typing import List, Dict
from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.PropertyPrivilege import PropertyPrivilege


class User(BaseModel):
    '''
        Data class for User.
    '''

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 createdTime: datetime = None,
                 graphPrivileges: Dict[str, List[str]] = None,
                 systemPrivileges: List[str] = None,
                 propertyPrivileges: PropertyPrivilege  = None,
                 policies: List[str] = None
                 ):
        self.username = username
        self.password = password
        self.createdTime = createdTime
        self.graphPrivileges = graphPrivileges
        self.systemPrivileges = systemPrivileges
        self.policies = policies
        self.propertyPrivileges = propertyPrivileges



from ultipa import DBType
from ultipa.operate.base_extra import BaseExtra
from ultipa.types.types_response import *
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.utils.convert import convertTableToDict


class LteUfeExtra(BaseExtra):

	'''
	Processsing class that defines settings for LTE and UFE related operations.
	'''

	def lte(self, dbType:DBType, propertyName:str,schemaName:str=None,
			config: RequestConfig = RequestConfig()) -> JobResponse:
		'''
		Load properties to memory (LTE).

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			propertyName: The name of property

			config: An object of RequestConfig class

		Returns:
			JobResponse
		'''

		command = dbType == DBType.DBNODE and CommandList.lteNode or CommandList.lteEdge
		uqlMaker = UQLMAKER(command=command, commonParams=config)
		if schemaName:
			commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		elif schemaName==None:
			commandP="@`*`.`%s`" %(propertyName)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		result = [{'new_job_id': ''}]
		if res.status.code == ErrorCode.SUCCESS and res.items:
			result = convertTableToDict(res.alias(res.aliases[0].alias).entities.rows,
										res.alias(res.aliases[0].alias).entities.headers)
		return JobResponse(jobId=result[0].get('new_job_id'), status=res.status, statistics=res.statistics)

	def ufe(self, dbType:DBType, propertyName:str, schemaName:str=None,
			config: RequestConfig = RequestConfig()) -> JobResponse:
		'''
		Unload properties from memory (UFE).

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			propertyName: The name of property

			config: An object of RequestConfig class

		Returns:
			JobResponse
		'''

		command = dbType == DBType.DBNODE and CommandList.ufeNode or CommandList.ufeEdge
		uqlMaker = UQLMAKER(command=command, commonParams=config)
		if schemaName:
			commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		elif schemaName==None:
			commandP="@`*`.`%s`" %(propertyName)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		result = [{'new_job_id': ''}]
		if res.status.code == ErrorCode.SUCCESS and res.items:
			result = convertTableToDict(res.alias(res.aliases[0].alias).entities.rows,
										res.alias(res.aliases[0].alias).entities.headers)
		return JobResponse(jobId=result[0].get('new_job_id'), status=res.status, statistics=res.statistics)



# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Index.py
from typing import List

from ultipa.structs import DBType
from ultipa.structs.BaseModel import BaseModel


class Process(BaseModel):
    '''
        Data class for Index.
    '''

    def __init__(self,
                 processId: str = None,
                 processQuery: str = None,
                 duration: str = None,
                 status: str = None):
        self.processId = processId
        self.processQuery = processQuery
        self.duration = duration
        self.status = status



from ultipa.types.types_response import Privilege

from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA
from ultipa.types.types_response import *
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.errors import ParameterException
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from typing import List
from ultipa.structs.Policy import Policy
from ultipa.utils.convert import convertToPolicy
from ultipa.utils.convert import convertToPrivilege
from ultipa.utils.noneCheck import checkNone

JSONSTRING_KEYS = ["graphPrivileges", "systemPrivileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class PolicyExtra(BaseExtra):
    '''
        Processing class that defines settings for policy related operations.
    '''

    def showPolicy(self,
                   config: RequestConfig = RequestConfig()) -> List[Policy]:
        '''
        Show policy list.

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Policy]

        '''
        uqlMaker = UQLMAKER(command=CommandList.showPolicy, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker,
                                 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
        if len(res.data) > 0:
            res.data = convertToPolicy(res)
        return res.data

    def showPrivilege(self,
                      config: RequestConfig = RequestConfig()) -> List[Privilege]:
        '''
        Show privilege list.

        Args:
            config: An object of RequestConfig class

        Returns:
            List[Privilege]

        '''

        uqlMaker = UQLMAKER(command=CommandList.showPrivilege, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker,
                                 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS))
        if len(res.data) > 0:
            res.data = convertToPrivilege(res)
        return res.data

    def getPolicy(self, policyName: str,
                  config: RequestConfig = RequestConfig()) -> Policy:
        '''
        Get a policy.

        Args:
            policyName: The name of policy

            config: An object of RequestConfig class

        Returns:
            Policy

        '''

        uqlMaker = UQLMAKER(command=CommandList.getPolicy, commonParams=config)
        uqlMaker.setCommandParams(policyName)
        res = self.UqlListSimple(uqlMaker=uqlMaker,
                                 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
        if res.status.code == ULTIPA.ErrorCode.SUCCESS and res.data:
            res.data = convertToPolicy(res)
        return res.data

    def createPolicy(self, policy: Policy,
                     config: RequestConfig = RequestConfig()) -> Response:
        '''
        Create a policy.

        Args:
            policy: The policy to be created; the field name must be set, systemPrivileges, graphPrivileges, propertyPrivilege and policies are optional.

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        if checkNone(policy.name):
            return Response(status=Status(code=ErrorCode.PARAM_ERROR,message='The name of policy cannot be None'))
        uqlMaker = UQLMAKER(command=CommandList.createPolicy, commonParams=config)
        uqlMaker.setCommandParams(policy.name)
        paramsP = []
        if policy.graphPrivileges:
            paramsP.append(f"graph_privileges:{policy.graphPrivileges}")

        if policy.systemPrivileges:
            paramsP.append(f"system_privileges:{policy.systemPrivileges}")

        if policy.policies:
            paramsP.append(f"policies:{policy.policies}")

        if policy.propertyPrivileges:
            paramsP.append(f"property_privileges:{policy.propertyPrivileges.to_dict()}")

        result = "{"+",".join(paramsP)+"}"
        uqlMaker.addParam('params',result,notQuotes = True)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        return res

    def alterPolicy(self, policy: Policy,
                    config: RequestConfig = RequestConfig()) -> Response:
        '''
        Alter a policy.

        Args:
            policy: The policy to be created; the field name must be set, systemPrivileges, graphPrivileges, propertyPrivilege and policies are optional.

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        if checkNone(policy.name) :
            return Response(status=Status(code=ErrorCode.PARAM_ERROR,message='The name of policy cannot be None'))
        uqlMaker = UQLMAKER(command=CommandList.alterPolicy, commonParams=config)
        uqlMaker.setCommandParams(policy.name)
        params = {}
        if policy.systemPrivileges is not None:
            params.update({"system_privileges": policy.systemPrivileges})
        if policy.graphPrivileges is not None:
            params.update({"graph_privileges": policy.graphPrivileges})

        if policy.policies is not None:
            params.update({"policies": policy.policies})

        if policy.propertyPrivileges is not None:
            params.update({"property_privileges": policy.propertyPrivileges.to_dict()})

        uqlMaker.addParam('set', params, notQuotes=True)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        return res

    def dropPolicy(self, policyName: str,
                   config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop a policy.

        Args:
            policyName:  The name of policy to be dropped

            config: An object of RequestConfig class

        Returns:
            Response

        '''

        uqlMaker = UQLMAKER(command=CommandList.dropPolicy, commonParams=config)
        uqlMaker.setCommandParams(policyName)
        res = self.uqlSingle(uqlMaker=uqlMaker)
        return res



import json
from typing import List, Callable

from ultipa.types.types_response import *
from ultipa.operate.base_extra import BaseExtra
from ultipa.proto import ultipa_pb2
from ultipa.types import ULTIPA
from ultipa.utils.format import FormatType
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.connection.clientType import ClientType
from ultipa.operate.task_extra import TaskExtra

class DownloadExtra(BaseExtra):
	'''
	Processing class that defines settings for file downloading operation.

	'''

	def _downloadAlgoResultFile(self, fileName: str,  cb: Callable[[bytes], None],
								config: RequestConfig = RequestConfig()):
		'''
		Download file.

		Args:
			fileName: Name of the file

			cb: Callback function that accepts bytes

			config: An object of RequestConfig class

		'''
		downResponse = UltipaResponse()
		try:

			clientInfo = self.getClientInfo(graphSetName=config.graph,
											clientType=ClientType.Leader)
			res = clientInfo.Controlsclient.DownloadFile(
				ultipa_pb2.DownloadFileRequest(file_name=fileName), metadata=clientInfo.metadata)

			for data_flow in res:
				ultipa_response = UltipaResponse()
				status = FormatType.status(data_flow.status)
				ultipa_response.status = status
				if status.code != ULTIPA.ErrorCode.SUCCESS:
					cb(ultipa_response)
					break
				ultipa_response.data = data_flow.chunk
				cb(ultipa_response.data)
		except Exception as e:
			downResponse.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=str(e))
			print(downResponse.status.message)

	def downloadAlgoResultFile(self, fileName: str,  cb: Callable[[bytes], None],
							   config: RequestConfig = RequestConfig()):
		'''
		Download file.

		Args:
			fileName: Name of the file

			cb:Callback function that accepts bytes

			config: An object of RequestConfig class

		'''
		return self._downloadAlgoResultFile(fileName=fileName, cb=cb, config=config)
	def _downloadAllAlgoResultFile(self, fileName: str,  cb: Callable[[bytes, str], None],
								   config: RequestConfig = RequestConfig()):
		'''
		Download all files.

		Args:
			fileName: Name of the file

			cb: Callback function that accepts bytes and string inputs

			config: An object of RequestConfig class

		'''
		downResponse = UltipaResponse()
		try:

			clientInfo = self.getClientInfo(graphSetName=config.graph,
											clientType=ClientType.Leader)
			res = clientInfo.Controlsclient.DownloadFile(
				ultipa_pb2.DownloadFileRequest(file_name=fileName), metadata=clientInfo.metadata)

			for data_flow in res:
				ultipa_response = UltipaResponse()
				status = FormatType.status(data_flow.status)
				ultipa_response.status = status
				if status.code != ULTIPA.ErrorCode.SUCCESS:
					cb(ultipa_response)
					break
				ultipa_response.data = data_flow.chunk
				cb(ultipa_response.data, fileName)
		except Exception as e:
			downResponse.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=str(e))
			print(downResponse.status.message)

	def downloadAllAlgoResultFile(self, cb: Callable[[bytes, str], None],jobId: str=None,
								  config: RequestConfig = RequestConfig()):
		'''
		Download all files.

		Args:

			jobId: id of the jobId

			cb: callback function for receiving data

			config: An object of RequestConfig class

		'''
		result = TaskExtra.showJob(self, id=jobId)
		filename = []
		if result:
			for data in result:
				if data.id == str(jobId):
					for key,value in json.loads(data.result).items():
						if key[:11] == 'output_file':
							filename.append(value)

		else:
			raise Exception('Job not found')

		for name in filename:
			self._downloadAllAlgoResultFile(fileName=name, cb=cb, config=config)



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 17:17
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : base_extra.py
import copy
import csv
import json
import types

from ultipa.structs.InsertErrorCode import  InsertErrorCodeMap
from ultipa.types.types import CodeMap
from ultipa.types.types_response import *
from ultipa import ParameterException
from ultipa.configuration import InsertRequestConfig
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.connection.clientType import ClientType
from ultipa.connection.commonUql import GetPropertyBySchema
from ultipa.connection.connectionBase import ConnectionBase
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.proto import ultipa_pb2
from ultipa.structs import DBType, Node, Edge
from ultipa.structs.Retry import Retry
from ultipa.structs.Schema import Schema
from ultipa.structs.Stats import Stats
from ultipa.types import ULTIPA, ULTIPA_REQUEST
from ultipa.types.types_response import PropertyTable, QueryResponseListener
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertTableToDict, convertToListAnyObject, convertToStats
from ultipa.utils.format import FormatType
from ultipa.utils.noneCheck import checkNone
from ultipa.utils.raftRetry import RetryHelp
from ultipa.utils.ultipa_datetime import getTimeZoneOffset, getTimeOffsetSeconds
from ultipa.structs.QLType import QLType

class BaseExtra(ConnectionBase):
	'''
		Processing class that defines settings for basic operations.

	'''

	def test(self,
			 config: RequestConfig = RequestConfig()) -> Response:
		'''
		Test connection.

		Args:
			config: An object of RequestConfig class

		Returns:
			Response
		'''
		testResponse = Response()
		returnReq = ULTIPA.ReturnReq(config.graph, "test", None, None, False)
		try:
			clientInfo = self.getClientInfo(useHost=config.host)
			name = 'Test'
			res = clientInfo.Controlsclient.SayHello(ultipa_pb2.HelloUltipaRequest(name=name),
													 metadata=clientInfo.metadata)
			returnReq.host = clientInfo.host
			if (res.message == name + " Welcome To Ultipa!"):

				testResponse.status = ULTIPA.Status(code=CodeMap.getCode(res.status.error_code), message=res.status.msg)
		except Exception as e:
			testResponse = Response()
			try:
				message = str(e._state.details)
			except:
				message = str(e)
			testResponse.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)
		return testResponse


	def exportData(self, request: ULTIPA_REQUEST.ExportRequest, cb: Callable[[List[Node], List[Edge]],None],
				   config: RequestConfig = RequestConfig()):
		try:
			req = ultipa_pb2.ExportRequest(db_type=request.dbType.value, limit=-1,
										   select_properties=request.selectProperties, schema=request.schema)

			graphName = request.graph if not checkNone(request.graph) else config.graph
			clientInfo = self.getClientInfo(graphSetName=graphName)
			res = clientInfo.Controlsclient.Export(req, metadata=clientInfo.metadata)
			nodedata = []
			edgedata = []
			response = UltipaResponse()
			for exportReply in res:
				response.status = FormatType.status(exportReply.status)
				if exportReply.node_table:
					nodedata = FormatType.export_nodes(exportReply, config.timeZone, config.timeZoneOffset)
				if exportReply.edge_table:
					edgedata = FormatType.export_edges(exportReply, config.timeZone, config.timeZoneOffset)
				if nodedata:
					uql = ULTIPA.ExportReply(data=nodedata)
					response.data = uql.data
					cb(uql.data,None)
				if edgedata:
					uql = ULTIPA.ExportReply(data=edgedata)
					response.data = uql.data
					cb(None,uql.data)

		except Exception as e:
			errorRes = UltipaResponse()
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
				print(message)
			except:
				message = 'UNKNOW ERROR'
				print(message)
			errorRes.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)
			return errorRes

	def uql(self, uql: str,
			config: RequestConfig = RequestConfig()) -> Response:
		'''
		Execute UQL.

		Args:
			uql: A uql statement

			config: An object of RequestConfig class

		Returns:
			Response

		'''
		request = ultipa_pb2.QueryRequest()
		request.query_text = uql
		request.query_type = QLType.UQL
		request.timeout = self.getTimeout(config.timeout)
		if config.thread is not None:
			request.thread_num = config.thread
		ultipaRes = Response()
		if config.graph == '' and self.defaultConfig.defaultGraph != '':
			config.graph = self.defaultConfig.defaultGraph

		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(Retry())
		canRetry=True
		while onRetry.current < onRetry.max and canRetry:
			try:
				import pytz
				getTimeZoneOffset(config, self.defaultConfig)
				timeZone = config.timeZone
				timeZoneOffset = config.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=config.graph, uql=uql,
												useHost=config.host,
												timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				uqlIsExtra = UQLHelper.uqlIsExtra(uql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.QueryEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Query(request, metadata=clientInfo.metadata)

				ultipaRes = FormatType.uqlMergeResponse(res, timeZone, timeZoneOffset)


				if not isinstance(ultipaRes, types.GeneratorType) and RetryHelp.checkRes(ultipaRes):
					onRetry.current += 1
					continue
				else:
					return ultipaRes

			except Exception as e:
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)
				if ultipaRes.status.code not in {ULTIPA.ErrorCode.RAFT_REDIRECT, ULTIPA.ErrorCode.RAFT_LEADER_NOT_YET_ELECTED, ULTIPA.ErrorCode.RAFT_NO_AVAILABLE_FOLLOWERS,ULTIPA.ErrorCode.RAFT_NO_AVAILABLE_ALGO_SERVERS}:
					canRetry=False
				else:
					onRetry.current += 1
					self.hostManagerControl.getHostManger(config.graph).raftReady = False


		return ultipaRes

	def uqlStream(self, uql: str, cb: QueryResponseListener, config: RequestConfig = RequestConfig()):

		'''
		Execute UQL.

		Args:
			uql: A uql statement

			cb: Listener for the streaming process.

			config: An object of RequestConfig class

		'''
		cb.emit("start", config)
		request = ultipa_pb2.QueryRequest()
		request.query_text = uql
		request.query_type = QLType.UQL
		request.timeout = self.getTimeout(config.timeout)
		if config.thread is not None:
			request.thread_num = config.thread


		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(Retry())
		while onRetry.current < onRetry.max:
			try:
				import pytz
				getTimeZoneOffset(config, self.defaultConfig)
				timeZone = config.timeZone
				timeZoneOffset = config.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=config.graph, uql=uql,
												useHost=config.host,
												 timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				uqlIsExtra = UQLHelper.uqlIsExtra(uql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.QueryEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Query(request, metadata=clientInfo.metadata)

				uql_response = UltipaResponse()
				ultipa_response = Response()
				for uqlReply in res:
					status = FormatType.status(uqlReply.status)
					uql_response = FormatType.response(uql_response, uqlReply, timeZone, timeZoneOffset)
					ret = ULTIPA.UqlReply(dataBase=uql_response.data)

					if status.code != ULTIPA.ErrorCode.SUCCESS:
						ultipa_response.status = uql_response.status
						cb.emit("end", config)
						return

					ultipa_response.items = ret._aliasMap
					ultipa_response.status = uql_response.status
					ultipa_response.statistics = uql_response.statistics
					should_continue = cb.emit("data", ultipa_response, config)
					if should_continue == False:
						cb.emit("end", config)
						return

				if not isinstance(ultipa_response, types.GeneratorType) and RetryHelp.checkRes(ultipa_response):
					onRetry.current += 1
					continue
				else:
					cb.emit("end", config)
					return

			except Exception as e:
				ultipaRes = Response()
				onRetry.current += 1
				self.hostManagerControl.getHostManger(config.graph).raftReady = False
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)
				print(ultipaRes.status.message)
		cb.emit("end", config)
		return

	def gql(self, gql: str,
			config: RequestConfig = RequestConfig()) -> Response:
		'''
		Execute GQL.

		Args:
			gql: A gql statement

			config: An object of RequestConfig class

		Returns:
			Response

		'''
		request = ultipa_pb2.QueryRequest()
		request.query_text = gql
		request.query_type = QLType.GQL
		request.timeout = self.getTimeout(config.timeout)
		if config.thread is not None:
			request.thread_num = config.thread
		ultipaRes = Response()
		if config.graph == '' and self.defaultConfig.defaultGraph != '':
			config.graph = self.defaultConfig.defaultGraph


		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(Retry())
		canRetry = True
		while onRetry.current < onRetry.max and canRetry:
			try:
				import pytz
				getTimeZoneOffset(config, self.defaultConfig)
				timeZone = config.timeZone
				timeZoneOffset = config.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=config.graph, uql=gql,
												useHost=config.host,
												timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				uqlIsExtra = UQLHelper.uqlIsExtra(gql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.QueryEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Query(request, metadata=clientInfo.metadata)


				ultipaRes = FormatType.uqlMergeResponse(res, timeZone, timeZoneOffset)

				if not isinstance(ultipaRes, types.GeneratorType) and RetryHelp.checkRes(ultipaRes):
					onRetry.current += 1
					continue
				else:
					return ultipaRes

			except Exception as e:
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)

				if ultipaRes.status.code not in {ULTIPA.ErrorCode.RAFT_REDIRECT, ULTIPA.ErrorCode.RAFT_LEADER_NOT_YET_ELECTED,
												 ULTIPA.ErrorCode.RAFT_NO_AVAILABLE_FOLLOWERS,
												 ULTIPA.ErrorCode.RAFT_NO_AVAILABLE_ALGO_SERVERS}:
					canRetry = False
				else:
					onRetry.current += 1
					self.hostManagerControl.getHostManger(config.graph).raftReady = False

		return ultipaRes

	def gqlStream(self, gql: str, cb: QueryResponseListener, config: RequestConfig = RequestConfig()):

		'''
		Execute UQL.

		Args:
			gql: A gql statement

			cb: Listener for the streaming process.

			config: An object of RequestConfig class

		'''
		cb.emit("start", config)
		request = ultipa_pb2.QueryRequest()
		request.query_text = gql
		request.query_type = QLType.GQL
		request.timeout = self.getTimeout(config.timeout)
		if config.thread is not None:
			request.thread_num = config.thread
		if config.graph == '' and self.defaultConfig.defaultGraph != '':
			config.graph = self.defaultConfig.defaultGraph


		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(Retry())
		while onRetry.current < onRetry.max:
			try:
				import pytz
				getTimeZoneOffset(config, self.defaultConfig)
				timeZone = config.timeZone
				timeZoneOffset = config.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=config.graph, uql=gql,
												useHost=config.host,
												timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				uqlIsExtra = UQLHelper.uqlIsExtra(gql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.QueryEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Query(request, metadata=clientInfo.metadata)

				uql_response = UltipaResponse()
				ultipa_response = Response()
				for uqlReply in res:
					status = FormatType.status(uqlReply.status)
					uql_response = FormatType.response(uql_response, uqlReply, timeZone, timeZoneOffset)
					ret = ULTIPA.UqlReply(dataBase=uql_response.data)

					if status.code != ULTIPA.ErrorCode.SUCCESS:
						ultipa_response.status = uql_response.status
						cb.emit("end", config)
						return

					ultipa_response.items = ret._aliasMap
					ultipa_response.status = uql_response.status
					ultipa_response.statistics = uql_response.statistics
					should_continue = cb.emit("data", ultipa_response, config)
					if should_continue == False:
						cb.emit("end", config)
						return

				if not isinstance(ultipa_response, types.GeneratorType) and RetryHelp.checkRes(ultipa_response):
					onRetry.current += 1
					continue
				else:
					cb.emit("end", config)
					return

			except Exception as e:
				ultipaRes = Response()
				onRetry.current += 1
				self.hostManagerControl.getHostManger(config.graph).raftReady = False
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)
				print(ultipaRes.status.message)
		cb.emit("end", config)
		return



	def uqlSingle(self, uqlMaker: UQLMAKER) -> Response:
		res = self.uql(uqlMaker.toString(), uqlMaker.commonParams)
		return res

	def UqlListSimple(self, uqlMaker: UQLMAKER, responseKeyFormat: ResponseKeyFormat = None,
					  isSingleOne: bool = True) -> Response:
		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.ErrorCode.SUCCESS:
			simplrRes = UltipaResponse(res.status, res.items)
			return simplrRes

		if not isSingleOne:
			retList = []
			for alias in res.aliases:
				item = res.items.get(alias.alias)
				table = item.entities
				table_rows = table.rows
				table_rows_dict = convertTableToDict(table_rows, table.headers)
				if responseKeyFormat:
					table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
				data = convertToListAnyObject(table_rows_dict)
				retList.append(PropertyTable(name=table.name, data=data))
			simplrRes = UltipaResponse(res.status, retList)
			return simplrRes

		alisFirst = res.aliases[0].alias if len(res.aliases) > 0 else None
		firstItem = res.items.get(alisFirst)
		if firstItem:
			table_rows = firstItem.entities.rows
			table_rows_dict = convertTableToDict(table_rows, firstItem.entities.headers)
			if responseKeyFormat:
				table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
			data = convertToListAnyObject(table_rows_dict)
			simplrRes = UltipaResponse(res.status, data)
			simplrRes.statistics = res.statistics
			return simplrRes
		else:
			return res

	def UqlUpdateSimple(self, uqlMaker: UQLMAKER):
		res = self.uqlSingle(uqlMaker)
		return UltipaResponse(res.status, statistics=res.statistics)

	def insertNodesBatchBySchema(self, schema: Schema, nodes: List[Node],
								 config: InsertRequestConfig) -> InsertResponse:
		'''
		Batch insert nodes of a same schema (that already exists in the graph).

		Args:
			schema:  The target schema; the attributes name and dbType are mandatory, properties should include some or all properties.

			nodes: The data to be inserted, List[Node]

			config: An object of InsertConfig class

		Returns:
			InsertResponse

		'''

		if checkNone(schema.name) or not isinstance(schema.properties,list) :
			messege = 'The name of schema cannot be None' if checkNone(schema.name) else 'The properties type of schema must be list'
			return Response(status=Status(code=ErrorCode.PARAM_ERROR,message=messege))

		if config.graph == '' and self.defaultConfig.defaultGraph != '':
			config.graph = self.defaultConfig.defaultGraph

		clientInfo = self.getClientInfo(clientType=ClientType.Update, graphSetName=config.graph
										)

		nodetable = FormatType.makeEntityNodeTable(schema, nodes, getTimeZoneOffset(requestConfig=config,
																				   defaultConfig=self.defaultConfig))

		_nodeTable = ultipa_pb2.EntityTable(schemas=nodetable.schemas, entity_rows=nodetable.nodeRows)
		request = ultipa_pb2.InsertNodesRequest()
		request.silent = config.silent
		request.insert_type = config.insertType.value
		request.graph_name = config.graph
		request.node_table.MergeFrom(_nodeTable)
		res = clientInfo.Rpcsclient.InsertNodes(request, metadata=clientInfo.metadata)

		status = FormatType.status(res.status)
		uqlres = InsertResponse(status=status)
		reTry = RetryHelp.check(self, config, uqlres)
		if reTry.canRetry:
			return self.insertNodesBatchBySchema(schema, nodes, config)

		uqlres.uuids = [i for i in res.uuids]
		uqlres.ids = [i for i in res.ids]
		errorDict = {}
		for i, data in enumerate(res.ignore_error_code):
			try:
				index = nodes[res.ignore_indexes[i]]._getIndex()
			except Exception as e:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			if index is None:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			errorDict.update({index: InsertErrorCodeMap.getUnsertErrorCode(data)})
		uqlres.errorItems = {key :errorDict[key] for key in sorted(errorDict.keys())}
		return uqlres

	def insertEdgesBatchBySchema(self, schema: Schema, edges: List[Edge],
								 config: InsertRequestConfig) -> InsertResponse:
		'''
		Batch insert edges of a same schema (that already exists in the graph)

		Args:
			schema: The target schema; the attributes name and dbType are mandatory, properties should include some or all properties.

			edges: The list of edges to be inserted; the attributes fromId and toId of each Edge are mandatory, uuid, fromUuid, and toUuid cannot be set, values must have the same structure with Schema.properties.

			config: An object of InsertConfig class

		Returns:
			InsertResponse

		'''
		if checkNone(schema.name) or not isinstance(schema.properties,list) :
			messege = 'The name of schema cannot be None' if checkNone(schema.name) else 'The properties type of schema must be list'
			return Response(status=Status(code=ErrorCode.PARAM_ERROR, message=messege))
		if config.graph == '' and self.defaultConfig.defaultGraph != '':
			config.graph = self.defaultConfig.defaultGraph

		clientInfo = self.getClientInfo(clientType=ClientType.Update, graphSetName=config.graph
										)

		edgetable = FormatType.makeEntityEdgeTable(schema=schema, rows = edges,
												   timeZoneOffset=getTimeZoneOffset(requestConfig=config,
																					defaultConfig=self.defaultConfig))

		_edgeTable = ultipa_pb2.EntityTable(schemas=edgetable.schemas, entity_rows=edgetable.edgeRows)
		request = ultipa_pb2.InsertEdgesRequest()
		request.silent = config.silent
		request.insert_type = config.insertType.value
		request.graph_name = config.graph
		request.edge_table.MergeFrom(_edgeTable)
		res = clientInfo.Rpcsclient.InsertEdges(request, metadata=clientInfo.metadata)

		status = FormatType.status(res.status)
		uqlres = InsertResponse(status=status)
		reTry = RetryHelp.check(self, config, uqlres)
		if reTry.canRetry:
			return self.insertEdgesBatchBySchema(schema, edges, config)

		uqlres.uuids = [i for i in res.uuids]
		errorDict = {}
		for i, data in enumerate(res.ignore_error_code):
			try:
				index = edges[res.ignore_indexes[i]]._getIndex()
			except Exception as e:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			if index is None:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			errorDict.update({index: InsertErrorCodeMap.getUnsertErrorCode(data)})
		uqlres.errorItems = {key :errorDict[key] for key in sorted(errorDict.keys())}
		return uqlres

	def insertNodesBatchAuto(self, nodes: List[Node],
							 config: InsertRequestConfig) -> Dict[str,InsertResponse]:
		'''
		Batch insert nodes of different schemas (that will be created if not existent)

		Args:
			nodes: The data to be inserted, List[Node]

			config: An object of InsertConfig class

		Returns:
			Dict[str,InsertResponse]

		'''
		Result = {}
		schemaDict = {}
		batches = {}
		schemaRet = self.uql(GetPropertyBySchema.node, config)
		if schemaRet.status.code == ULTIPA.ErrorCode.SUCCESS:
			for aliase in schemaRet.aliases:
				if aliase.alias == '_nodeSchema':
					schemaDict = convertTableToDict(schemaRet.alias(aliase.alias).entities.rows,
													schemaRet.alias(aliase.alias).entities.headers)
			if not schemaDict:
				raise ParameterException(err='Please create Node Schema.')
		else:
			raise ParameterException(err=schemaRet.status.message)
		for index, node in enumerate(nodes):
			node._index = index
			if batches.get(node.schema) is None:
				batches[node.schema] = ULTIPA_REQUEST.Batch()
				find = list(filter(lambda x: x.get('name') == node.schema, schemaDict))
				if find:
					findSchema = find[0]
					propertyList = FormatType.checkProperty(node, json.loads(findSchema.get("properties")))
					reqSchema = ULTIPA_REQUEST.Schema(name=node.schema, properties=propertyList,dbType=DBType.DBNODE)
					batches[node.schema].Schema = reqSchema
				else:
					if node.schema is None:
						raise ParameterException(err=f"Row [{index}]:Please set schema name for node.")
					else:
						raise ParameterException(err=f"Row [{index}]:Node Schema not found {node.schema}.")

			batches.get(node.schema).Nodes.append(node)
		for key in batches:
			batch = batches.get(key)
			ret = self.insertNodesBatchBySchema(schema=batch.Schema, nodes=batch.Nodes, config=config)
			Result.update(
				{key:ret })
		return Result


	def insertEdgesBatchAuto(self, edges: List[Edge],
							 config: InsertRequestConfig) -> Dict[str,InsertResponse]:
		'''
		Batch insert edges of different schemas (that will be created if not existent)

		Args:
			edges: The data to be inserted, List[Edge]

			config: An object of InsertConfig class

		Returns:
			Dict[str,InsertResponse]

		'''
		Result = {}
		schemaDict = []
		batches = {}
		schemaRet = self.uql(GetPropertyBySchema.edge, config)
		if schemaRet.status.code == ULTIPA.ErrorCode.SUCCESS:
			for aliase in schemaRet.aliases:
				if aliase.alias == '_edgeSchema':
					schemaDict = convertTableToDict(schemaRet.alias(aliase.alias).entities.rows,
													schemaRet.alias(aliase.alias).entities.headers)
			if not schemaDict:
				raise ParameterException(err='Please create Edge Schema.')
		else:
			raise ParameterException(err=schemaRet.status.message)
		for index, edge in enumerate(edges):
			edge._index = index
			if batches.get(edge.schema) == None:
				batches[edge.schema] = ULTIPA_REQUEST.Batch()
				find = list(filter(lambda x: x.get('name') == edge.schema, schemaDict))
				if find:
					findSchema = find[0]
					propertyList = FormatType.checkProperty(edge, json.loads(findSchema.get("properties")))
					reqSchema = ULTIPA_REQUEST.Schema(name=edge.schema, properties=propertyList,dbType=DBType.DBEDGE)
					batches[edge.schema].Schema = reqSchema
				else:
					if edge.schema is None:
						raise ParameterException(err=f"Row [{index}]:Please set schema name for edge.")
					else:
						raise ParameterException(err=f"Row [{index}]:Edge Schema not found {edge.schema}.")
			batches.get(edge.schema).Edges.append(edge)
		for key in batches:
			batch = batches.get(key)
			ret = self.insertEdgesBatchBySchema(schema=batch.Schema, edges=batch.Edges, config=config)
			Result.update(
				{key:ret})
		return Result



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:40
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : PropertyType.py
from enum import Enum

from ultipa.proto import ultipa_pb2


class PropertyTypeStr:
    '''
        Data class for property type mapping to string.
    '''
    INT = 'int32'
    STRING = 'string'
    FLOAT = 'float'
    DOUBLE = 'double'
    UINT32 = 'uint32'
    INT64 = 'int64'
    UINT64 = 'uint64'
    DATETIME = 'datetime'
    TIMESTAMP = 'timestamp'
    TEXT = 'text'
    BLOB = "blob"
    BOOL = "bool"
    UNSET = "unset"
    POINT = "point"
    LIST = "list"
    SET = "set"
    MAP = "map"
    BLOB = "blob"


class UltipaPropertyType(Enum):
    '''
        Data class for property type mapping to gRPC.
    '''
    UNSET = ultipa_pb2.UNSET
    INT32 = ultipa_pb2.INT32
    STRING = ultipa_pb2.STRING
    FLOAT = ultipa_pb2.FLOAT
    DOUBLE = ultipa_pb2.DOUBLE
    UINT32 = ultipa_pb2.UINT32
    INT64 = ultipa_pb2.INT64
    UINT64 = ultipa_pb2.UINT64
    DATETIME = ultipa_pb2.DATETIME
    TIMESTAMP = ultipa_pb2.TIMESTAMP
    TEXT = ultipa_pb2.TEXT
    BLOB = ultipa_pb2.BLOB
    BOOL = ultipa_pb2.BOOL
    POINT = ultipa_pb2.POINT
    DECIMAL = ultipa_pb2.DECIMAL
    LIST = ultipa_pb2.LIST
    SET = ultipa_pb2.SET
    MAP = ultipa_pb2.MAP
    NULL = ultipa_pb2.NULL_
    ID = -1
    UUID = -2
    FROM = -3
    TO = -4
    FROM_UUID = -5
    TO_UUID = -6
    IGNORE = -7





from ultipa.operate.base_extra import BaseExtra
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.format import FormatType
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.connection.clientType import ClientType
from ultipa.types import ULTIPA, ULTIPA_REQUEST
from ultipa.types.types import *
from ultipa.utils.convert import convertToUser
from ultipa.types.types_response import *
from ultipa.utils.noneCheck import checkNone

JSONSTRING_KEYS = ["graphPrivileges", "systemPrivileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class UserExtra(BaseExtra):
    '''
        Processing class that defines settings for user related operations.
    '''

    def _GRPATH_PRIVILEGES_DATA_FORMAT(self, obj):
        if isinstance(obj.get('graph_privileges'), list):
            resr = FormatType.graphPrivileges(obj.get('graph_privileges'))
            return resr
        else:
            return '[]'

    def showUser(self,
                 config: RequestConfig = RequestConfig()) -> List[User]:
        '''
        Show user list.

        Args:
            config: An object of RequestConfig class

        Returns:
            List[User]

        '''

        uqlMaker = UQLMAKER(command=CommandList.showUser, commonParams=config)
        res = self.UqlListSimple(uqlMaker=uqlMaker,
                                 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
        if len(res.data) > 0:
            res.data = convertToUser(res)
        else:
            res.data = None
        return res.data

    def getUser(self, username: str,
                config: RequestConfig = RequestConfig()) -> User:
        '''
        Get a designated user.

        Args:
            username: The name of user

            config: An object of RequestConfig class

        Returns:
            User

        '''

        uqlMaker = UQLMAKER(command=CommandList.getUser, commonParams=config)
        uqlMaker.setCommandParams(username)
        res = self.UqlListSimple(uqlMaker=uqlMaker,
                                 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
        # if isinstance(res.data, list) and len(res.data) > 0:
        # 	res.data = res.data[0]
        if res.status.code == ULTIPA.ErrorCode.SUCCESS and res.data:
            res.data = convertToUser(res)
        return res.data

    def createUser(self, user: User,
                   config: RequestConfig = RequestConfig()) -> Response:
        '''
        Create a user.

        Args:
            user: The user to be created; the fields username and password must be set, systemPrivileges, graphPrivileges, propertyPrivilege and policies are optional

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        if checkNone(user.username) or checkNone(user.password):
            return Response(status=Status(code= ErrorCode.PARAM_ERROR,message='The user`s username and password cannot be None'))

        uqlMaker = UQLMAKER(command=CommandList.createUser, commonParams=config)
        userparam = []

        if user.username:
            userparam.append(user.username)
        else:
            raise ParameterException(err='username is a required parameter')

        if user.password:
            userparam.append(user.password)
        else:
            raise ParameterException(err='password is a required parameter')

        uqlMaker.setCommandParams(userparam)

        params = []

        if user.graphPrivileges:
            params.append(f"graph_privileges:{user.graphPrivileges}")

        if user.systemPrivileges:
            params.append(f"system_privileges:{user.systemPrivileges}")

        if user.policies:
            params.append(f"policies:{user.policies}")

        if user.propertyPrivileges:
            params.append(f"property_privileges:{user.propertyPrivileges.to_dict()}")

        result = "{" + ",".join(params) + "}"
        uqlMaker.addParam('params', result, notQuotes=True)

        return self.uqlSingle(uqlMaker=uqlMaker)

    def dropUser(self, username: str,
                 config: RequestConfig = RequestConfig()) -> Response:
        '''
        Drop a user.

        Args:
            username: The name of user

            config: An object of RequestConfig class

        Returns:
            Response

        '''

        uqlMaker = UQLMAKER(command=CommandList.dropUser, commonParams=config)
        uqlMaker.setCommandParams(username)
        return self.uqlSingle(uqlMaker=uqlMaker)

    def alterUser(self, user: User,
                  config: RequestConfig = RequestConfig()) -> Response:
        '''
        Alter a user.

        Args:
            user: The user to be altered; the fields username must be set, systemPrivileges, graphPrivileges, propertyPrivilege and policies are optional.

            config: An object of RequestConfig class

        Returns:
            Response

        '''
        if checkNone(user.username):
            return Response(status=Status(code= ErrorCode.PARAM_ERROR,message='The user`s username cannot be None'))
        uqlMaker = UQLMAKER(command=CommandList.alterUser, commonParams=config)
        if user.username:
            uqlMaker.setCommandParams(user.username)
        else:
            raise ParameterException(err='username is a required parameter')

        paramsDict = {}
        if user.password:
            paramsDict.setdefault('password', user.password)

        if user.graphPrivileges:
            paramsDict.setdefault('graph_privileges', user.graphPrivileges)

        if user.systemPrivileges:
            paramsDict.setdefault('system_privileges', user.systemPrivileges)

        if user.policies:
            paramsDict.setdefault('policies', user.policies)

        if user.propertyPrivileges:
            paramsDict.setdefault('property_privileges', user.propertyPrivileges.to_dict())

        uqlMaker.addParam('set', paramsDict)
        return self.uqlSingle(uqlMaker=uqlMaker)



# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Index.py
from typing import List

from ultipa.structs import DBType
from ultipa.structs.BaseModel import BaseModel


class Index(BaseModel):
    '''
        Data class for Index.
    '''

    def __init__(self,
                 name: str = None,
                 properties: str = None,
                 schema: str = None,
                 status: str = None,
                 size: str = None,
                 id: str = None,
                 dbType: DBType = None
                 ):
        self.id = id
        self.name = name
        self.properties = properties
        self.schema = schema
        self.status = status
        self.size = size
        self.dbType = dbType



# -*- coding: utf-8 -*-

import threading
import time
import queue


class TooManyConnections(Exception):
	...


class Expired(Exception):
	...

class UsageExceeded(Expired):
	...

class TtlExceeded(Expired):
	...

class IdleExceeded(Expired):
	...

class WrapperConnection(object):

	def __init__(self, pool, connection):
		self.pool = pool
		self.connection = connection
		self.usage = 0
		self.last = self.created = time.time()

	def using(self):
		self.usage += 1
		self.last = time.time()
		return self

	def reset(self):
		self.usage = self.last = self.created = 0

	def __enter__(self):
		return self.connection

	def __exit__(self, *exc_info):
		self.pool.release(self)


class ConnectionPool():
	'''
		Connection pool

	'''

	__wrappers = {}

	def __init__(self, create, close=None, max_size=3, max_usage=0, ttl=0, idle=120, block=True):
		if not hasattr(create, '__call__'):
			raise ValueError('"create" argument is not callable')

		if close is not None and not hasattr(close, '__call__'):
			raise ValueError('"close" argument is not callable')

		self._create = create
		self._close = close
		self._max_size = int(max_size)
		self._max_usage = int(max_usage)
		self._ttl = int(ttl)
		self._idle = int(idle)
		self._block = bool(block)
		self._lock = threading.Condition()
		self._pool = queue.Queue()
		self._size = 0
		if self._pool.qsize() == 0:
			while self._max_size:
				self._pool.put_nowait(self._wrapper(self._create()))
				self._max_size -= 1

	def item(self):
		self._lock.acquire()

		try:
			while (self._max_size and self._pool.empty() and self._size >= self._max_size):
				if not self._block:
					raise TooManyConnections('Too many connections')

				self._lock.wait()  # Waiting for idle connections
			if self._pool.qsize() == 0:
				while self._max_size:
					self._pool.put_nowait(self._wrapper(self._create()))
					self._max_size -= 1
			try:
				wrapped = self._pool.get_nowait()  # Get an idle connection from the connection pool
				if self._idle and (wrapped.last + self._idle) < time.time():
					self._destroy(wrapped, isCreate=True)
					raise IdleExceeded('Idle exceeds %d secs' % self._idle)
			except (queue.Empty, IdleExceeded):
				wrapped = self._wrapper(self._create())  # Create new connection
				self._size += 1
		finally:
			self._lock.release()

		return wrapped.using()

	def release(self, conn):
		self._lock.acquire()
		wrapped = self._wrapper(conn)

		try:
			self._test(wrapped)
		except Expired:
			self._destroy(wrapped)
		else:
			self._pool.put_nowait(wrapped)
			self._lock.notifyAll()  # Notify the rest of threads on the availability of idle connection
		finally:
			self._lock.release()

	def _destroy(self, wrapped, isCreate: bool = False):
		if self._close:
			self._close(wrapped.connection)

		self._unwrapper(wrapped)
		self._size += 1
		if isCreate:
			self._pool.put_nowait(self._wrapper(self._create()))
			self._size -= 1

	def _wrapper(self, conn):
		if isinstance(conn, WrapperConnection):
			return conn

		_id = id(conn)

		if _id not in self.__wrappers:
			self.__wrappers[_id] = WrapperConnection(self, conn)

		return self.__wrappers[_id]

	def _unwrapper(self, wrapped):
		if not isinstance(wrapped, WrapperConnection):
			return

		_id = id(wrapped.connection)
		wrapped.reset()
		del wrapped

		if _id in self.__wrappers:
			del self.__wrappers[_id]

	def _test(self, wrapped):
		if self._max_usage and wrapped.usage >= self._max_usage:
			raise UsageExceeded('Usage exceeds %d times' % self._max_usage)

		if self._ttl and (wrapped.created + self._ttl) < time.time():
			raise TtlExceeded('TTL exceeds %d secs' % self._ttl)



import hashlib
import os
from typing import List


from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.proto import ultipa_pb2
from ultipa.types import ULTIPA_REQUEST, ULTIPA
from ultipa.types.types_response import *
from ultipa.types.types import Exta
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.fileSize import read_in_chunks
from ultipa.utils.format import FormatType
from ultipa.utils.convert import  convertToExta


class AlgoExtra(BaseExtra):
	'''
		Processing class that defines settings for algorithm related operations.

	'''
	JSONSTRING_KEYS = ["param"]


	def __make_message(self, fileName, md5, chunk, hdcName):
		return ultipa_pb2.InstallAlgoRequest(
			file_name=fileName, md5=md5, chunk=chunk, with_server=ultipa_pb2.WithServer(hdc_server_name=hdcName)
		)

	def __generate_messages(self, request: ULTIPA_REQUEST.InstallAlgo):
		messages = []
		hdcName = request.hdcName
		for file in request.files:
			file_object = open(file, 'rb')
			filemd5 = hashlib.md5(file_object.read()).hexdigest()
			file_object.close()

			for chunk in read_in_chunks(file):
				fileName = os.path.basename(file)
				messages.append(self.__make_message(fileName, filemd5, chunk, hdcName))
		for msg in messages:
			yield msg


	def installHDCAlgo(self, files:List[str],hdcServerName:str,
					config: RequestConfig = RequestConfig()) -> Response:
		'''
		Installs an Ultipa graph algorithm in the instance.

		Args:
			files(List[str]): 1、The directory of the algorithm package (.so)
							  2、The directory of the algorithm configuration (.yml)

			hdcServerName: The name of the hdc

			config: An object of the RequestConfig class

		Returns:
			Response

		'''
		request = ULTIPA_REQUEST.InstallAlgo(files,hdcServerName)
		clientInfo = self.getClientInfo(graphSetName=config.graph,
										isGlobal=True)
		response = UltipaResponse()
		try:
			if len(files)>0 :
				installRet = clientInfo.Controlsclient.InstallAlgo(self.__generate_messages(request),
																   metadata=clientInfo.metadata)
				status = FormatType.status(installRet.status)
				response.status = status
		except Exception as e:
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = str(e)

			response.status = ULTIPA.Status(code=ULTIPA.ErrorCode.UNKNOW_ERROR, message=message)

		return response

	def uninstallHDCAlgo(self, algoName: str,hdcServerName:str,
					  config: RequestConfig = RequestConfig()) -> Response:
		'''
		Uninstall an Ultipa graph algorithm in the instance.

		Args:
			algoName: The name of algorithm

			hdcServerName: The name of hdc

			config: An object of the RequestConfig class

		Returns:
			Response

		'''
		clientInfo = self.getClientInfo(graphSetName=config.graph,
										isGlobal=True)
		arequest = ultipa_pb2.UninstallAlgoRequest(algo_name=algoName, with_server=ultipa_pb2.WithServer(hdc_server_name=hdcServerName))
		installRet = clientInfo.Controlsclient.UninstallAlgo(arequest, metadata=clientInfo.metadata)
		status = FormatType.status(installRet.status)
		response = Response(status=status)
		return response

	def rollbackHDCAlgo(self,algoName :str,hdcServerName:str
						,config: RequestConfig = RequestConfig()) -> Response:
		"""
		Args:
			algoName: The name of algorithm

			hdcServerName: The name of hdc

			config: An object of the RequestConfig class

		Returns:
			Response
		"""
		clientInfo = self.getClientInfo(graphSetName=config.graph,
										isGlobal=True)
		arequest = ultipa_pb2.RollbackAlgoRequest(algo_name=algoName,
												   with_server=ultipa_pb2.WithServer(hdc_server_name=hdcServerName))
		rollbackRet = clientInfo.Controlsclient.RollbackAlgo(arequest, metadata=clientInfo.metadata)
		status = FormatType.status(rollbackRet.status)
		response = Response(status=status)
		return response

	def showHDCAlgo(self,hdcServerName:str
						,config: RequestConfig = RequestConfig()) -> List[Algo]:
		"""
		Args:
			hdcServerName: The name of hdc

			config: An object of the RequestConfig class

		Returns:
			Response
		"""
		command=CommandList.showHDCAlgo
		commandp=f"'{hdcServerName}'"
		uqlMaker = UQLMAKER(command=command, commandP=commandp, commonParams=config)
		res=self.uqlSingle(uqlMaker)
		result = []
		if res.status.code == ULTIPA.ErrorCode.SUCCESS:
			if res.alias(f"_algoList_from_{hdcServerName}").entities:
				result = res.alias(f"_algoList_from_{hdcServerName}").asAlgos()



		return result



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:21
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : EntityRow.py
from typing import Dict


class EntityRow:
    '''
        Data class for data rows (nodes or edges) to be inserted.
    '''
    _index = None

    def __init__(self, values: Dict,
                 schema: str = None,
                 id: str = None,
                 fromId: str = None, toId: str = None,
                 uuid: int = None, fromUuid: int = None,
                 toUuid: int = None):
        self.uuid = uuid
        self.id = id
        self.fromUuid = fromUuid
        self.toUuid = toUuid
        self.fromId = fromId
        self.toId = toId
        self.schema = schema
        self.values = values

    def _getIndex(self):
        return self._index



# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 14:46
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Property.py
from typing import List, Dict

from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.PropertyType import UltipaPropertyType


class Property(BaseModel):
    '''
        Data class for property.
    '''

    @staticmethod
    def DecimalExtra(accuracy, scale):
        return f"{accuracy},{scale}"

    def __init__(self,
                 name: str = None,
                 type: UltipaPropertyType = None,
                 subType: List[UltipaPropertyType] = None,
                 lte: bool = None,
                 read: bool = None,
                 write: bool = None,
                 schema: str = None,
                 description: str = None,
                 encrypt: str = None,
                 extra: str = None
                 ):
        self.type = type
        self.subType = subType
        self.description = description
        self.name = name
        self.lte = lte
        self.schema = schema
        self.encrypt = encrypt
        self.extra = extra
        self.read = read
        self.write = write

    _PropertyMap = {
        "string": UltipaPropertyType.STRING,
        "int32": UltipaPropertyType.INT32,
        "int64": UltipaPropertyType.INT64,
        "uint32": UltipaPropertyType.UINT32,
        "uint64": UltipaPropertyType.UINT64,
        "float": UltipaPropertyType.FLOAT,
        "double": UltipaPropertyType.DOUBLE,
        "datetime": UltipaPropertyType.DATETIME,
        "timestamp": UltipaPropertyType.TIMESTAMP,
        "text": UltipaPropertyType.TEXT,
        "blob": UltipaPropertyType.BLOB,
        "_id": UltipaPropertyType.ID,
        "_uuid": UltipaPropertyType.UUID,
        "_from": UltipaPropertyType.FROM,
        "_to": UltipaPropertyType.TO,
        "_from_uuid": UltipaPropertyType.FROM_UUID,
        "_to_uuid": UltipaPropertyType.TO_UUID,
        "_ignore": UltipaPropertyType.IGNORE,
        "unset": UltipaPropertyType.UNSET,
        "point": UltipaPropertyType.POINT,
        "decimal": UltipaPropertyType.DECIMAL,
        "list": UltipaPropertyType.LIST,
        "set": UltipaPropertyType.SET,
        "map": UltipaPropertyType.MAP,
        "null": UltipaPropertyType.NULL,
        "bool": UltipaPropertyType.BOOL
    }

    _PropertyReverseMap = {
        UltipaPropertyType.STRING: "string",
        UltipaPropertyType.INT32: "int32",
        UltipaPropertyType.INT64: "int64",
        UltipaPropertyType.UINT32: "uint32",
        UltipaPropertyType.UINT64: "uint64",
        UltipaPropertyType.FLOAT: "float",
        UltipaPropertyType.DOUBLE: "double",
        UltipaPropertyType.DATETIME: "datetime",
        UltipaPropertyType.TIMESTAMP: "timestamp",
        UltipaPropertyType.TEXT: "text",
        UltipaPropertyType.BLOB: "blob",
        UltipaPropertyType.ID: "_id",
        UltipaPropertyType.UUID: "_uuid",
        UltipaPropertyType.FROM: "_from",
        UltipaPropertyType.TO: "_to",
        UltipaPropertyType.FROM_UUID: "_from_uuid",
        UltipaPropertyType.TO_UUID: "_to_uuid",
        UltipaPropertyType.IGNORE: "_ignore",
        UltipaPropertyType.UNSET: "unset",
        UltipaPropertyType.POINT: "point",
        UltipaPropertyType.DECIMAL: "decimal",
        UltipaPropertyType.LIST: "list",
        UltipaPropertyType.SET: "set",
        UltipaPropertyType.MAP: "map",
        UltipaPropertyType.NULL: "null",
        UltipaPropertyType.BOOL: "bool",
    }

    _PropertyEnumMap = {
        UltipaPropertyType.STRING.value: UltipaPropertyType.STRING,
        UltipaPropertyType.INT32.value: UltipaPropertyType.INT32,
        UltipaPropertyType.INT64.value: UltipaPropertyType.INT64,
        UltipaPropertyType.UINT32.value: UltipaPropertyType.UINT32,
        UltipaPropertyType.UINT64.value: UltipaPropertyType.UINT64,
        UltipaPropertyType.FLOAT.value: UltipaPropertyType.FLOAT,
        UltipaPropertyType.DOUBLE.value: UltipaPropertyType.DOUBLE,
        UltipaPropertyType.DATETIME.value: UltipaPropertyType.DATETIME,
        UltipaPropertyType.TIMESTAMP.value: UltipaPropertyType.TIMESTAMP,
        UltipaPropertyType.TEXT.value: UltipaPropertyType.TEXT,
        UltipaPropertyType.BLOB.value: UltipaPropertyType.BLOB,
        UltipaPropertyType.ID.value: UltipaPropertyType.ID,
        UltipaPropertyType.UUID.value: UltipaPropertyType.UUID,
        UltipaPropertyType.FROM.value: UltipaPropertyType.FROM,
        UltipaPropertyType.TO.value: UltipaPropertyType.TO,
        UltipaPropertyType.FROM_UUID.value: UltipaPropertyType.FROM_UUID,
        UltipaPropertyType.TO_UUID.value: UltipaPropertyType.TO_UUID,
        UltipaPropertyType.IGNORE.value: UltipaPropertyType.IGNORE,
        UltipaPropertyType.UNSET.value: UltipaPropertyType.UNSET,
        UltipaPropertyType.POINT.value: UltipaPropertyType.POINT,
        UltipaPropertyType.DECIMAL.value: UltipaPropertyType.DECIMAL,
        UltipaPropertyType.LIST.value: UltipaPropertyType.LIST,
        UltipaPropertyType.SET.value: UltipaPropertyType.SET,
        UltipaPropertyType.MAP.value: UltipaPropertyType.MAP,
        UltipaPropertyType.NULL.value: UltipaPropertyType.NULL,
        UltipaPropertyType.BOOL.value: UltipaPropertyType.BOOL
    }

    def setSubTypesbyType(self, type: str):
        if "string" in type:
            self.subType = [UltipaPropertyType.STRING]

        if "int32" in type:
            self.subType = [UltipaPropertyType.INT32]

        if "uint32" in type:
            self.subType = [UltipaPropertyType.UINT32]

        if "int64" in type:
            self.subType = [UltipaPropertyType.INT64]

        if "uint64" in type:
            self.subType = [UltipaPropertyType.UINT64]

        if "float" in type:
            self.subType = [UltipaPropertyType.FLOAT]

        if "double" in type:
            self.subType = [UltipaPropertyType.DOUBLE]

        if "datetime" in type:
            self.subType = [UltipaPropertyType.DATETIME]

        if "timestamp" in type:
            self.subType = [UltipaPropertyType.TIMESTAMP]

        if "text" in type:
            self.subType = [UltipaPropertyType.TEXT]

    def isIdType(self) -> bool:
        idTypes = [
            UltipaPropertyType.ID.value,
            UltipaPropertyType.TO.value,
            UltipaPropertyType.UUID.value,
            UltipaPropertyType.FROM.value,
            UltipaPropertyType.FROM_UUID.value,
            UltipaPropertyType.TO_UUID.value,
        ]
        idTypesEnum = [
            UltipaPropertyType.ID,
            UltipaPropertyType.TO,
            UltipaPropertyType.UUID,
            UltipaPropertyType.FROM,
            UltipaPropertyType.FROM_UUID,
            UltipaPropertyType.TO_UUID,
        ]
        return self.type in idTypes or self.type in idTypesEnum

    def isIgnore(self):
        return self.type == UltipaPropertyType.IGNORE.value

    def setTypeStr(self, value):
        self.type = self.getStringByPropertyType(value)

    def setTypeInt(self, value):
        self.type = self.getPropertyTypeByString(value)

    def getStringType(self):
        return self.getStringByPropertyType(self.type)

    def getPropertyTypeByString(self, v):
        if isinstance(v, str) and 'decimal' in v:
            self.extra = v
            v = 'decimal'
        if not self._PropertyMap.get(v):
            if not self._PropertyReverseMap.get(v):
                if "[" in v:
                    self.setSubTypesbyType(v)
                    return UltipaPropertyType.LIST
                if '<' in v:
                    self.setSubTypesbyType(v)
                    return UltipaPropertyType.SET
            else:
                return v

        return self._PropertyMap.get(v)

    def getStringByPropertyType(self, v):
        if v == UltipaPropertyType.DECIMAL:
            return f'decimal({self.extra})'
        if isinstance(v,str):
            return v
        return self._PropertyReverseMap[v]
    @staticmethod
    def getPropertyByInt(v):
        if v:
            return Property._PropertyEnumMap[v]
        return v

    @staticmethod
    def _getStringByPropertyType(v):
        return Property._PropertyReverseMap[Property.getPropertyByInt(v)]

    @staticmethod
    def _getPropertyTypeByString(v):
        return Property._PropertyMap.get(v)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Graph.py
from typing import Dict, List

from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.Path import Path
from ultipa.structs.Node import Node
from ultipa.structs.Edge import Edge


class Graph(BaseModel):
    '''
        Data class for graph.
    '''

    def __init__(self,
                 paths: List[Path] = [],
                 nodes: Dict[int, Node] = {},
                 edges: Dict[int, Edge] = {}):
        self.paths = paths
        self.nodes = nodes
        self.edges = edges

    def getPaths(self):
        for data in self.paths:
            data.nodes = {nodeuuidsdata: self.nodes[nodeuuidsdata] for nodeuuidsdata in data.nodeUuids}
            data.edges = {edgeuuidsdata: self.edges[edgeuuidsdata] for edgeuuidsdata in data.edgeUuids}
        return self.paths

    def addNode(self, node: Node):
        self.nodes.update({node.uuid: node})

    def addEdge(self, edge: Edge):
        self.edges.update({edge.uuid: edge})


class GraphAlias(BaseModel):
    def __init__(self, alias: str, graph: Graph = None):
        self.alias = alias
        self.graph = graph



class ParameterException(Exception):
    '''
        Class for customizing errors related to parameter
    '''
    def __init__(self,err='Parameter error!'):
        Exception.__init__(self,err)


class ServerException(Exception):
    '''
       Class for customizing errors related to server connection
    '''
    def __init__(self,err='Server connection failed!'):
        Exception.__init__(self,err)


class SerializeException(Exception):
    '''
       Class for customizing errors related to serialization
    '''
    def __init__(self,err='Serialize failed!'):
        Exception.__init__(self,err)



class SettingException(Exception):
    '''
        Class for customizing errors related to settings
    '''
    def __init__(self,err='Setting error!'):
        Exception.__init__(self,err)


def checkError(error: str):
    '''
        Method for customizing errors related to parameter validation
    '''
    if "large" in error:
        return "argument out of range"
    else:
        return error